"""
This script contains the general implementation of the Task Arithmetic method.

http://arxiv.org/abs/2212.04089
"""

import functools
import logging
import os
from abc import abstractmethod
from collections import defaultdict
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, TypeVar, Union

import torch
from lightning.fabric.utilities.rank_zero import rank_zero_only
from omegaconf import DictConfig
from torch import Tensor, nn
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm

from fusion_bench.compat.method import ModelFusionAlgorithm
from fusion_bench.compat.modelpool import HuggingFaceClipVisionPool, ModelPool
from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.mixins.lightning_fabric import LightningFabricMixin
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.models.wrappers.layer_wise_fusion import (
    LayerWiseMergedModel,
    get_layer_wise_weights,
)
from fusion_bench.utils.data import load_tensor_from_file
from fusion_bench.utils.type import TorchModelType

from .utils import *

if TYPE_CHECKING:
    from fusion_bench.programs.fabric_fusion_program import FabricModelFusionProgram

from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils import instantiate
from fusion_bench.utils.data import InfiniteDataLoader
from fusion_bench.utils.state_dict_arithmetic import (
    state_dict_add,
    state_dict_mul,
    state_dict_sub,
)
from fusion_bench.utils.type import StateDictType

log = logging.getLogger(__name__)


@torch.no_grad()
def task_arithmetic_merge(
    pretrained_model: nn.Module,
    finetuned_models: List[Dict[str, Tensor]],
    scaling_factor: float,
    inplace: bool = True,
) -> nn.Module:
    """
    Merges the task vectors from multiple fine-tuned models into a single pre-trained model.

    Args:
        pretrained_model (nn.Module): The pre-trained model to which the task vectors will be added.
        finetuned_models (List[nn.Module]): A list of fine-tuned models from which task vectors will be calculated.
        scaling_factor (float): A factor by which the task vectors will be scaled before merging.
        inplace (bool, optional): If True, the pre-trained model will be modified in place.
                                  If False, a copy of the pre-trained model will be modified. Defaults to True.

    Returns:
        nn.Module: The pre-trained model with the merged task vectors.
    """
    if not inplace:
        pretrained_model = deepcopy(pretrained_model)
    if isinstance(finetuned_models[0], nn.Module):
        finetuned_models = [
            deepcopy(model.state_dict(keep_vars=True)) for model in finetuned_models
        ]
    task_vector: StateDictType = None
    # Calculate the total task vector
    for model in finetuned_models:
        if task_vector is None:
            task_vector = state_dict_sub(
                model,
                pretrained_model.state_dict(keep_vars=True),
            )
        else:
            task_vector = state_dict_add(
                task_vector,
                state_dict_sub(
                    model,
                    pretrained_model.state_dict(keep_vars=True),
                ),
            )
    # scale the task vector
    task_vector = state_dict_mul(task_vector, scaling_factor)
    # add the task vector to the pretrained model
    state_dict = state_dict_add(
        pretrained_model.state_dict(keep_vars=True), task_vector
    )
    pretrained_model.load_state_dict(state_dict)
    return pretrained_model


@torch.no_grad()
def ties_merge(
    pretrained_model: nn.Module,
    finetuned_models: List[Dict[str, Tensor]],
    scaling_factor: float,
    threshold: float,
) -> nn.Module:
    remove_keys = []
    merge_func = "sum"
    if isinstance(finetuned_models[0], nn.Module):
        finetuned_models = [
            deepcopy(model.state_dict(keep_vars=True)) for model in finetuned_models
        ]

    ptm_check = pretrained_model.state_dict(keep_vars=True)

    # Compute the task vectors
    flat_ft = torch.vstack(
        [state_dict_to_vector(check, remove_keys) for check in finetuned_models]
    )
    flat_ptm = state_dict_to_vector(ptm_check, remove_keys)
    tv_flat_checks = flat_ft - flat_ptm

    # Perform TIES Merging
    merged_tv = ties_merging(
        tv_flat_checks,
        reset_thresh=threshold,
        merge_func=merge_func,
    )
    merged_check = flat_ptm + scaling_factor * merged_tv
    merged_state_dict = vector_to_state_dict(
        merged_check, ptm_check, remove_keys=remove_keys
    )

    # Load the merged state dict into the pretrained model
    pretrained_model.load_state_dict(merged_state_dict)
    return pretrained_model


def entropy_loss(logits: Tensor, pred=None, eps: float = 1e-8) -> Tensor:
    """
    Compute the entropy loss of a set of logits.

    Args:
        logits (Tensor): The logits to compute the entropy loss of.
        eps (float): A small value to avoid log(0). Default is 1e-8.

    Returns:
        Tensor: The entropy loss of the logits.
    """
    # Ensure the logits tensor has 2 dimensions
    assert (
        logits.dim() == 2
    ), f"Expected logits to have 2 dimensions, found {logits.dim()}, {logits.size()=}"

    # Compute the softmax probabilities
    probs = torch.softmax(logits, dim=-1)

    # Compute the entropy loss
    return -torch.sum(probs * torch.log(probs + eps), dim=-1).mean()


class FrankWolfeHardAlgorithm(
    CLIPClassificationMixin,
    ModelFusionAlgorithm,
    SimpleProfilerMixin,
):

    def __init__(
        self,
        merge_fn: str,
        step_size: float,
        max_iters: int,
        dataset_size: int,
        tasks: List[str] = [],
        granularity: str = "task",
        max_num_models: int = 100,
        loss_fn: str = "cross_entropy",
        init_weight: str = "",
        scaling_factor: float = 1.0,
        threshold: int = 20,
        **kwargs,
    ):
        """
        Initializes the TaskArithmeticAlgorithm with the given scaling factor.

        Args:
            scaling_factor (int): The factor by which the task vectors will be scaled before merging.
        """
        self.merger = merge_fn
        if merge_fn == "task_arithmetic":
            self.merge_fn = task_arithmetic_merge
        elif merge_fn == "ties":
            self.merge_fn = partial(ties_merge, threshold=threshold)
        # elif merge_fn == "concrete_ta":
        #     self.merge_fn = ConcreteTaskArithmeticAlgorithmForCLIP(
        #         instantiate(OmegaConf.load("config/method/concrete_subspace/clip_concrete_task_arithmetic.yaml"))
        #     )
        else:
            raise ValueError(f"Unsupported merge_fn: {merge_fn}")
        self.scaling_factor = scaling_factor

        self.init_weight = init_weight
        self.step_size = step_size
        self.max_iters = max_iters
        self.granularity = granularity
        self.loss_fn = loss_fn
        self.tasks = tasks
        self.dataset_size = dataset_size
        self.max_num_models = max_num_models
        super().__init__(**kwargs)

    def on_frank_wolfe_iteration_start(self):
        self.setup_zero_shot_classification_head()

    @functools.cache
    def get_shuffled_loader_iter(self, task: str):
        if self.loss_fn == "cross_entropy":
            # get dataloader kwargs
            dataloader_kwargs = self.dataloader_kwargs.copy()
            dataloader_kwargs["shuffle"] = True
            dataloader_kwargs["batch_size"] = 1

            # get the test dataset
            clip_dataset = CLIPDataset(
                self.modelpool.load_train_dataset(task), self.clip_processor
            )
            # create the dataloader
            loader = DataLoader(clip_dataset, **dataloader_kwargs)
            loader = self.fabric.setup_dataloaders(loader)
            return iter(InfiniteDataLoader(loader))
        elif self.loss_fn == "entropy":
            return super().get_shuffled_test_loader_iter(
                task,
                batch_size=1,
            )
        else:
            raise ValueError(f"Unsupported loss function: {self.loss_fn}")

    def frank_wolfe_iteration(self, merged_model):

        merged_model.train()
        # zero the gradients
        for name, param in merged_model.named_parameters():
            param.requires_grad = True
            param.grad = None

        if self.loss_fn == "cross_entropy":
            loss_fn = nn.CrossEntropyLoss()
        elif self.loss_fn == "entropy":
            loss_fn = entropy_loss
        avg_loss = defaultdict(list)
        tasks = self.tasks if self.tasks else self.modelpool.model_names
        for task in tasks:
            log.info(f"Processing task {task}")
            for _ in range(self.dataset_size):
                with self.profile("data loading"):
                    batch = next(self.get_shuffled_loader_iter(task))
                with self.profile("forward pass"):
                    logits = self.compute_logits(merged_model, batch[0], task)
                    loss = loss_fn(logits, batch[1]) / (
                        self.dataset_size * len(self.modelpool.model_names)
                    )
                with self.profile("backward pass"):
                    # self.fabric.backward(loss, retain_graph=True)
                    loss.backward()
                avg_loss[task].append(loss.item())

        # calculate the loss
        avg_loss = {
            task: sum(losses) / len(losses) for task, losses in avg_loss.items()
        }
        log.info(
            f"Average Loss: {avg_loss}, Total Loss: {sum(avg_loss.values()) / len(avg_loss)}"
        )

        gradients = {
            name: param.grad.clone().to("cpu")
            for name, param in merged_model.named_parameters()
            if param.requires_grad
        }
        for name, param in merged_model.named_parameters():
            param.grad = None
        merged_model.eval()

        return gradients

    def frank_wolfe_selection(
        self, gradients, checkpoints, model_to_merge_names={}, type="task"
    ):
        assert type in [
            "task",
            "layer",
        ], f"Unsupported FW selection type: {type}, supported types are ['task', 'layer']"
        min_inner_product = float("inf")
        min_model = None
        min_model_name = None
        log_dict = {}
        if type == "task":
            for model_name, model_to_merge in checkpoints.items():
                model_to_merge = model_to_merge.to("cpu").state_dict()
                inner_product_sum = 0
                for param_name, param_value in model_to_merge.items():
                    # caclulate consine similarity
                    grad = gradients[param_name]
                    ckpt = model_to_merge[param_name]
                    param_alignment = torch.dot(grad.flatten(), ckpt.flatten()) / (
                        torch.norm(grad) * torch.norm(ckpt)
                    )
                    inner_product_sum += param_alignment
                log_dict[model_name] = inner_product_sum.item()
                if (
                    inner_product_sum < min_inner_product
                    and model_name not in model_to_merge_names
                ):
                    min_inner_product = inner_product_sum
                    min_model = deepcopy(model_to_merge)
                    min_model_name = model_name
        else:
            min_model = {}
            min_inner_product = {}
            min_idx = {}
            min_model_name = {}
            for model_name, model_to_merge in checkpoints.items():
                model_to_merge = model_to_merge.to("cpu").state_dict()
                for param_name, param_value in model_to_merge.items():
                    # caclulate consine similarity
                    grad = gradients[param_name]
                    ckpt = model_to_merge[param_name]
                    param_alignment = torch.dot(grad.flatten(), ckpt.flatten()) / (
                        torch.norm(grad) * torch.norm(ckpt)
                    )
                    if (
                        param_name not in min_inner_product
                        or param_alignment < min_inner_product[param_name]
                    ) and model_name not in model_to_merge_names[param_name]:
                        min_inner_product[param_name] = param_alignment
                        # if min_inner_product[param_name] < 0:
                        min_model[param_name] = param_value
                        min_idx[param_name] = model_name
                        min_model_name[param_name] = model_name
                        # else:
                        # min_model[param_name] = torch.zeros_like(param_value)
            min_inner_product = sum(min_inner_product.values())
            log_dict = {model_name: 0 for model_name in checkpoints.keys()}
            for k in min_idx.values():
                log_dict[k] += 1

        return min_model, min_model_name, min_inner_product, log_dict

    def run(self, modelpool: HuggingFaceClipVisionPool):
        log.info("Fusing models using FW merging.")
        self.modelpool = modelpool
        self.log_hyperparams(self.config)
        self.on_frank_wolfe_iteration_start()

        assert modelpool.has_pretrained, "Pretrained model is required."
        finetuned_models = {
            name: modelpool.load_model(name)
            for name in modelpool.model_names[: self.max_num_models]
        }
        pretrained_model = modelpool.load_model("_pretrained_")

        if self.init_weight:
            if self.init_weight == "base":
                log.info("Initializing the merged model with the base model")
                merged_model = pretrained_model
            else:
                log.info("Initializing the merged model with the initial weight")
                if isinstance(self.init_weight, str):
                    # self.config.weights is a path to a saved tensor
                    layer_wise_weight = load_tensor_from_file(self.init_weight)
                else:
                    raise ValueError(f"Unsupported weights format: {self.init_weight}")

                merged_model = LayerWiseMergedModel(
                    layer_wise_weight=layer_wise_weight,
                    pretrained_model=modelpool.load_model("_pretrained_"),
                    finetuned_models=list(finetuned_models.values()),
                    clamp_weights=False,
                    tie_weights=True,
                    strict=False,
                ).cuda()
                merged_model = merged_model.merge_and_unload()
        else:
            log.info("Initializing the merged model with merge function")
            merged_model = self.merge_fn(
                pretrained_model=modelpool.load_model("_pretrained_"),
                finetuned_models=list(finetuned_models.values()),
                scaling_factor=self.scaling_factor,
            ).cuda()
        # merged_model = self.fabric.setup(merged_model)

        initial_model = modelpool.load_model("_pretrained_")
        initial_model.load_state_dict(deepcopy(merged_model.state_dict()))
        finetuned_models["initial"] = initial_model
        for step_idx in (
            pbar := tqdm(
                range(self.max_iters if not self.is_debug_mode else 1),
                ("[DEBUG MODE] " if self.is_debug_mode else "") + "Frank-Wolfe Merging",
                dynamic_ncols=True,
            )
        ):
            torch.cuda.empty_cache()
            torch.set_grad_enabled(True)
            gradients = self.frank_wolfe_iteration(merged_model.cuda())
            torch.set_grad_enabled(False)
            grad_norm = torch.norm(
                torch.stack([torch.norm(g) for g in gradients.values()])
            )

            model_to_merge_names = (
                []
                if self.granularity == "task"
                else {name: [] for name in merged_model.state_dict().keys()}
            )
            min_model, min_model_name, min_alignment, chosen_model = (
                self.frank_wolfe_selection(
                    gradients,
                    finetuned_models,
                    model_to_merge_names=model_to_merge_names,
                    type=self.granularity,
                )
            )

            # Determine step size
            step = 2 / (step_idx + 2) * self.step_size

            # print iteration information
            log.info(
                f"Iteration {step_idx+1}, Task Vector: {min_model_name}, Gradient Norm: {grad_norm:.6f}, Inner Products: {min_alignment:.6f}, Chosen Model: {chosen_model}"
            )

            merged_model = self.merge_fn(
                pretrained_model=merged_model.to("cpu"),
                finetuned_models=[min_model],
                scaling_factor=step * self.scaling_factor,
            )

        torch.set_grad_enabled(False)
        merged_model = merged_model.cuda().eval()
        return merged_model



import logging
import random
from collections import OrderedDict
from copy import deepcopy
from typing import Dict, List, Literal, Optional, Tuple

import numpy as np
import torch
from scipy.stats import ortho_group
from torch import Tensor, nn

from fusion_bench.method.base_algorithm import BaseAlgorithm
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import BaseModelPool
from fusion_bench.utils.parameters import get_parameter_summary, human_readable
from fusion_bench.utils.state_dict_arithmetic import state_dict_avg
from fusion_bench.utils.type import StateDictType

log = logging.getLogger(__name__)


def cosine_similarity(tensor1: Tensor, tensor2: Tensor) -> float:
    if tensor1.shape != tensor2.shape:
        raise ValueError("Matrices must have the same shape")
    vec1 = tensor1.flatten()
    vec2 = tensor2.flatten()
    dot_product = torch.sum(vec1 * vec2)
    norm1 = torch.sqrt(torch.sum(vec1**2))
    norm2 = torch.sqrt(torch.sum(vec2**2))
    if norm1 == 0 or norm2 == 0:
        return 0.0
    return dot_product / (norm1 * norm2)


def svd_and_partition(
    A: torch.Tensor, num_chunks: int = 3
) -> Tuple[List[torch.Tensor], List[torch.Tensor], List[torch.Tensor]]:
    U, S, V = torch.svd(A)
    singular_values = len(S)
    chunk_size = singular_values // num_chunks
    U_chunks, S_chunks, V_chunks = [], [], []

    for i in range(num_chunks):
        start_idx = i * chunk_size
        end_idx = singular_values if i == num_chunks - 1 else start_idx + chunk_size

        U_chunks.append(U[:, start_idx:end_idx])
        S_chunks.append(S[start_idx:end_idx])
        V_chunks.append(V[:, start_idx:end_idx])

    return U_chunks, S_chunks, V_chunks


def compute_svd_subspace_similarity(
    ref: torch.Tensor, retrieval: torch.Tensor, num_chunks: int = 3
) -> List[dict]:
    if torch.cuda.is_available():
        ref = ref.cuda()
        retrieval = retrieval.cuda()
    U_chunks, S_chunks, V_chunks = svd_and_partition(ref, num_chunks)
    similarities = []
    for i in range(num_chunks):
        retrieval_approx = (
            U_chunks[i] @ U_chunks[i].T @ retrieval @ V_chunks[i] @ V_chunks[i].T
        )
        frob_sim = torch.norm(ref - retrieval_approx, p="fro").item() / ref.numel()
        cos_sim = cosine_similarity(ref, retrieval_approx)
        if isinstance(cos_sim, torch.Tensor):
            cos_sim = cos_sim.item()
        similarities.append(
            {
                "subspace": i + 1,
                "frobenius_similarity": frob_sim,
                "cosine_similarity": cos_sim,
            }
        )
    return similarities


def pairwise_cosine_similarity_matrix(tensors: List[torch.Tensor]) -> torch.Tensor:
    if torch.cuda.is_available():
        tensors = [tensor.cuda() for tensor in tensors]
    n = len(tensors)
    similarity_matrix = torch.zeros((n, n))
    for i in range(n):
        for j in range(n):
            similarity = cosine_similarity(tensors[i], tensors[j])
            similarity_matrix[i, j] = similarity.item()
    return similarity_matrix


def compare_models(
    state_dict1: StateDictType, state_dict2: StateDictType, target_layers=None
):
    results = {
        "layerwise_l2": {},
        "layerwise_cosine_similarity": {},
        "total_l2": None,
        "average_l2": None,
        "total_cosine_similarity": None,
        "average_cosine_similarity": None,
    }
    # Initialize lists to store flattened parameters
    params1_list = []
    params2_list = []

    keys1 = set(state_dict1.keys())
    keys2 = set(state_dict2.keys())
    # filter out layers that are not in target_layers
    if target_layers is not None:
        keys1 = keys1.intersection(target_layers)
        keys2 = keys2.intersection(target_layers)

    common_keys = keys1 & keys2
    if keys1 != keys2:
        print(
            "Warning: State dicts have different keys. Comparison will be made on common keys only."
        )
    num_layers = len(common_keys)

    for key in common_keys:
        tensor1 = state_dict1[key].float()
        tensor2 = state_dict2[key].float()

        # Compute L2 norm difference
        l2_diff = torch.norm(tensor1 - tensor2, p=2) / tensor1.numel()
        results["layerwise_l2"][key] = l2_diff.item()

        # Compute cosine similarity
        tensor1_flat = tensor1.reshape(-1)
        tensor2_flat = tensor2.reshape(-1)
        cos_sim = cosine_similarity(tensor1_flat, tensor2_flat).item()
        results["layerwise_cosine_similarity"][key] = cos_sim

        # Collect parameters for total metrics
        params1_list.append(tensor1_flat)
        params2_list.append(tensor2_flat)

    # Compute total metrics over all parameters
    if params1_list and params2_list:
        params1 = torch.cat(params1_list)
        params2 = torch.cat(params2_list)
        # Compute total L2 norm difference
        total_l2_difference = (
            torch.norm(params1 - params2, p=2).item() / params1.numel()
        )
        results["total_l2"] = total_l2_difference
        # Compute total cosine similarity
        total_cosine_similarity = cosine_similarity(params1, params2).item()
        results["total_cosine_similarity"] = total_cosine_similarity
    else:
        results["total_l2"] = None
        results["total_cosine_similarity"] = None

    # Compute average metrics
    if num_layers > 0:
        average_l2 = sum(results["layerwise_l2"].values()) / num_layers
        average_cosine_similarity = (
            sum(results["layerwise_cosine_similarity"].values()) / num_layers
        )
        results["average_l2"] = average_l2
        results["average_cosine_similarity"] = average_cosine_similarity
    else:
        results["average_l2"] = None
        results["average_cosine_similarity"] = None

    return results


class SuperposedAlgorithmBase(
    BaseAlgorithm,
    SimpleProfilerMixin,
):
    _config_mapping = BaseAlgorithm._config_mapping | {
        "mode": "mode",
        "target_layer": "target_layer",
        "random_seed": "random_seed",
        "different_across_layers": "different_across_layers",
        "joint_matrix_mode": "joint_matrix_mode",
        "rank": "rank",
        "random_components": "random_components",
        "shift_layers": "shift_layers",
        "absorber": "absorber",
        "debug": "debug",
        "ms_mode": "ms_mode",
        "verbose": "verbose",
        "dropout_rate": "dropout_rate",
    }

    def __init__(
        self,
        mode: str,
        target_layer: str,
        random_seed: int,
        different_across_layers: bool,
        joint_matrix_mode: str,
        rank: int,
        random_components: bool,
        shift_layers: int,
        absorber: Literal["average", "pretrained", "None"],
        debug: int,
        ms_mode: str,
        verbose: int,
        dropout_rate: int,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.mode = mode
        self.target_layer = target_layer
        self.random_seed = random_seed
        self.different_across_layers = different_across_layers
        self.joint_matrix_mode = joint_matrix_mode
        self.rank = rank
        self.random_components = random_components
        self.shift_layers = shift_layers
        self.absorber = absorber
        self.debug = debug
        self.ms_mode = ms_mode
        self.verbose = verbose
        self.dropout_rate = dropout_rate

    def _compute_svd_subspace_similarities(
        self,
        original_state_dict: StateDictType,
        retrieved_state_dict: StateDictType,
        target_layers: Optional[List[str]] = None,
    ) -> dict:
        svd_similarities = {}
        for layer_name, original_param in original_state_dict.items():
            if target_layers is not None and layer_name not in target_layers:
                continue
            if (
                original_param.dim() == 2
            ):  # Only compute for 2D tensors (weight matrices)
                retrieved_param = retrieved_state_dict[layer_name]
                svd_similarities[layer_name] = compute_svd_subspace_similarity(
                    original_param.float(), retrieved_param.float()
                )
        return svd_similarities

    def _load_state_dicts(self, modelpool: BaseModelPool) -> Dict[str, StateDictType]:
        """
        Load the state dicts of the models in the modelpool.

        Args:
            modelpool (BaseModelPool): The modelpool to load the state dicts from.

        Returns:
            Dict[str, StateDictType]: A dictionary of state dicts, keyed by model name.
        """
        state_dicts = {}
        for model_name in modelpool.model_names:
            with self.profile("load model"):
                model = modelpool.load_model(model_name)
            state_dicts[model_name] = model.state_dict(keep_vars=True)
        return state_dicts

    def _compute_absorber(
        self,
        state_dicts: Dict[str, StateDictType],
        pretrained_model: Optional[nn.Module] = None,
    ) -> Optional[StateDictType]:
        """
        Compute the absorber state dict.

        Args:
            state_dicts (Dict[str, StateDictType]): The state dicts of the models, keyed by model name, i.e. `{model_name: state_dict}`.
            pretrained_model (Optional[nn.Module]): The pretrained model.

        Returns:
            Optional[StateDictType]: The absorber state dict.
        """
        if self.absorber == "average":
            return state_dict_avg(list(state_dicts.values()))
        elif self.absorber == "pretrained":
            return pretrained_model.state_dict(keep_vars=True)
        elif self.absorber == "None":
            return None
        else:
            raise ValueError(
                f"Unsupported absorber type: {self.absorber}. Must be one of 'average', 'pretrained', or 'None'."
            )

    @staticmethod
    def svd_decomposition(A, r):
        if torch.cuda.is_available():
            A = A.cuda()
        U, S, V = torch.svd(A)
        return (U[:, :r] @ torch.diag(S[:r])).cpu(), V.t()[:r, :].cpu()

    @staticmethod
    def svd_decomposition_bm(A, r_most, r_mid):
        if torch.cuda.is_available():
            A = A.cuda()

        # Perform SVD
        U, S, V = torch.svd(A)

        # Get the most significant 'r_most' dimensions
        U_most = U[:, :r_most]
        S_most = S[:r_most]
        V_most = V[:, :r_most]

        # Get the middle 'r_mid' dimensions
        start_mid = len(S) // 2 - r_mid // 2
        end_mid = start_mid + r_mid
        U_mid = U[:, start_mid:end_mid]
        S_mid = S[start_mid:end_mid]
        V_mid = V[:, start_mid:end_mid]

        # Combine the results into two sets
        U_combined = torch.cat([U_most, U_mid], dim=1)
        S_combined = torch.cat([S_most, S_mid])
        V_combined = torch.cat([V_most, V_mid], dim=1)

        return (U_combined @ torch.diag(S_combined)).cpu(), V_combined.t().cpu()

    @staticmethod
    def svd_decomposition(A, r=None, r_most=None, r_mid=None, random_components=False):
        """
        Perform SVD decomposition with options for:
        1. Truncated SVD with 'r' components (if r is provided and random_components=False).
        2. Most significant 'r_most' and middle 'r_mid' components (if r_most and r_mid are provided).
        3. Randomly selected 'r' components (if r is provided and random_components=True).

        Args:
            A (torch.Tensor): The input matrix to decompose.
            r (int, optional): Number of components for standard or random SVD.
            r_most (int, optional): Number of most significant components.
            r_mid (int, optional): Number of middle components.
            random_components (bool, optional): Whether to sample 'r' random components.

        Returns:
            (torch.Tensor, torch.Tensor): Two matrices resulting from the SVD decomposition.
        """
        if torch.cuda.is_available():
            A = A.cuda()

        # Perform SVD
        U, S, V = torch.svd(A)

        if r is not None and not random_components:
            # Standard SVD decomposition with 'r' components
            return (U[:, :r] @ torch.diag(S[:r])).cpu(), V.t()[:r, :].cpu()

        elif r_most is not None and r_mid is not None:
            # SVD decomposition with 'r_most' most significant and 'r_mid' middle components
            # Most significant components
            U_most = U[:, :r_most]
            S_most = S[:r_most]
            V_most = V[:, :r_most]

            # Middle components
            start_mid = len(S) // 2 - r_mid // 2
            end_mid = start_mid + r_mid
            U_mid = U[:, start_mid:end_mid]
            S_mid = S[start_mid:end_mid]
            V_mid = V[:, start_mid:end_mid]

            # Combine the most and middle components
            U_combined = torch.cat([U_most, U_mid], dim=1)
            S_combined = torch.cat([S_most, S_mid])
            V_combined = torch.cat([V_most, V_mid], dim=1)

            return (U_combined @ torch.diag(S_combined)).cpu(), V_combined.t().cpu()

        elif r is not None and random_components:
            # SVD decomposition with random 'r' components
            indices = torch.randperm(len(S))[:r]
            U_rand = U[:, indices]
            S_rand = S[indices]
            V_rand = V[:, indices]

            return (U_rand @ torch.diag(S_rand)).cpu(), V_rand.t().cpu()

        else:
            raise ValueError(
                "Invalid combination of arguments. Provide correct parameters."
            )

    @staticmethod
    def _get_rank(A, rank):
        if isinstance(rank, str):
            r1, r2 = rank.split("-")
            r1 = int(float(r1) * min(A.shape)) if "." in r1 else int(r1)
            r2 = int(float(r2) * min(A.shape)) if "." in r2 else int(r2)
            return r1, r2
        if isinstance(rank, int):
            return rank
        elif isinstance(rank, float):
            return int(rank * min(A.shape))

    def _target_layer_flag(self, layer: str):
        """
        The method takes a layer name as input and returns a boolean indicating whether this layer should be targeted.

        Current implementation assume Transformer architecture and layer number is the first number in the layer name.

        Args:
            layer (str): The name of the layer.

        Returns:
            bool: True if the layer should be targeted, False otherwise.
        """
        target_layers = self.target_layer  # e.g. ["mlp_w", "attn_w"]
        # TODO: figure out what wo is in flan-t5
        mlp_flag = "mlp" in layer or "Dense" in layer
        attn_flag = "attn" in layer or "Attention" in layer
        weight_flag = "weight" in layer
        bias_flag = "bias" in layer
        target_flags = []
        for target_layer in target_layers:
            if target_layer == "mlp_w":
                target_flags.append(mlp_flag and not bias_flag)
            elif target_layer == "attn_w":
                target_flags.append(attn_flag and not bias_flag)
            elif target_layer == "all":
                target_flags.append(True)
            elif target_layer == "mlp":
                target_flags.append(mlp_flag)
            elif target_layer == "attn":
                target_flags.append(attn_flag)
            else:
                raise ValueError(f"Unsupported target layer: {target_layer}")
        target_flag = any(target_flags)
        return target_flag

    def _compress_and_retrieve(self, state_dicts: Dict[str, StateDictType], mode: str):
        """
        Compress and retrieve the state dicts.

        Args:
            state_dicts (Dict[str, StateDictType]): The state dicts of the models, keyed by model name, i.e. `{model_name: state_dict}`.
            mode (str): The mode of the compression and retrieval.

        Returns:
            Dict[str, StateDictType]: The compressed and retrieved state dicts, keyed by model name, i.e. `{model_name: state_dict}`.
        """
        # Assume the state_dicts have the same layers.
        layers = state_dicts[list(state_dicts.keys())[0]].keys()
        models = list(state_dicts.keys())
        compressed_layers = {}
        compression_context = {model: {} for model in models}
        retrieval_context = {model: {} for model in models}
        retrieval_models = deepcopy(state_dicts)
        # target_layer_flags = [self._target_layer_flag(layer) for layer in layers]
        # implement target_layer_flags with dropout
        target_layer_flags: List[bool] = []
        count = 0
        for layer in layers:
            if self._target_layer_flag(layer):
                # take the target layer per `self.dropout_rate` target layers.
                # e.g. if self.dropout_rate = 2, then take the 2nd and 4th target layers, skip the first and third target layers.
                # if self.dropout_rate = 1, then take all target layers.
                count += 1
                if count == self.dropout_rate:
                    target_layer_flags.append(True)
                    count = 0
                else:
                    target_layer_flags.append(False)
            else:
                target_layer_flags.append(False)

        target_layers = [
            layer for layer, flag in zip(layers, target_layer_flags) if flag
        ]
        log.info(
            f"filtered {len(target_layers)} target layers out of {len(layers)} layers"
        )

        metadata = {
            "nonzero_parameter_count": 0,
            "nonzero_param_count_context": 0,
            "task_vector_retrieval_similarity": {},
            "superposed_model_retrieval_similarity": {},
            "model_retrieval_similarity": {},
            "target_layers": target_layers,
            "task_vector_svd_subspace_similarities": {},
            "superposed_model_svd_subspace_similarities": {},
            "model_svd_subspace_similarities": {},
            "total_param_count_original": 0,
            "total_gb_original": 0,
            "total_gb_retrieved": 0,
        }

        if "absorber" in models:
            models.remove("absorber")
            absorber = state_dicts["absorber"]
        else:
            absorber = None

        # get the total number of parameters and bytes (in GB) of the original model
        original_param_summary = get_parameter_summary(state_dicts[models[0]])
        gbs = original_param_summary["bytes"] / 1e9
        log.info(
            f"Total parameters: {human_readable(original_param_summary['all_param'])}"
        )
        log.info(f"Total gigabytes: {gbs}")
        metadata["total_param_count_original"] = original_param_summary["all_param"]
        metadata["total_gb_original"] = gbs

        # for analysis purposes
        if self.debug >= 2:
            test_models = models
            # test_models = models[:2]
            # layers_old = {model: OrderedDict() for model in models}
            layers_old = {model: deepcopy(state_dicts[model]) for model in models}
            tv_new = {
                model: {model: OrderedDict() for model in models}
                for model in test_models
            }
            # layers_new = {model: {model: OrderedDict() for model in models} for model in test_models}

        # Shift the layers
        # TODO: make this more robust to other models.
        if self.shift_layers != 0:
            # random shuffling. Do not shuffle layers with no number in their name.
            # because they are likely to be special layers like text embeddings.
            if self.shift_layers == -1:
                layer_mappings = {model: {} for model in models}
                temp_state_dicts = deepcopy(state_dicts)

                # get layer number index, assume the first number in the layer name is the layer number
                # assume all numbered layers have their number at the same index
                # assume components separated by '.' in the layer name
                found_digit = False
                for layer_idx, layer in enumerate(layers):
                    if target_layer_flags[layer_idx]:
                        layer_parts = layer.split(".")
                        for i, part in enumerate(layer_parts):
                            if part.isdigit():
                                layer_number_idx = i
                                break
                        if found_digit:
                            break

                # get groups of target layers with same name except the layer number
                target_layer_groups = {}
                for layer_idx, layer in enumerate(layers):
                    if target_layer_flags[layer_idx]:
                        layer_parts = layer.split(".")
                        if (
                            layer_number_idx >= len(layer_parts)
                            or not layer_parts[layer_number_idx].isdigit()
                        ):
                            continue  # skip layers without number
                        base_name = ".".join(
                            layer_parts[:layer_number_idx]
                            + layer_parts[layer_number_idx + 1 :]
                        )
                        layer_number = int(layer_parts[layer_number_idx])
                        if base_name not in target_layer_groups:
                            target_layer_groups[base_name] = []
                        target_layer_groups[base_name].append(layer_number)

                # construct random shuffled mapping
                random_shuffle_mapping = {model: {} for model in models}
                for model_idx, model in enumerate(models):
                    for layer_idx, layer in enumerate(layers):
                        if target_layer_flags[layer_idx]:
                            layer_parts = layer.split(".")
                            if (
                                layer_number_idx >= len(layer_parts)
                                or not layer_parts[layer_number_idx].isdigit()
                            ):
                                continue  # skip layers without number
                            base_name = ".".join(
                                layer_parts[:layer_number_idx]
                                + layer_parts[layer_number_idx + 1 :]
                            )

                            if base_name not in random_shuffle_mapping[model]:
                                rng_state = random.getstate()
                                # Shuffle the layer numbers differently for each model
                                random.seed(self.config.random_seed + model_idx)
                                shuffled_layer_numbers = target_layer_groups[
                                    base_name
                                ].copy()
                                random.shuffle(shuffled_layer_numbers)
                                random_shuffle_mapping[model][base_name] = {
                                    orig: str(shuffled)
                                    for orig, shuffled in zip(
                                        target_layer_groups[base_name],
                                        shuffled_layer_numbers,
                                    )
                                }
                                random.setstate(rng_state)

                    for layer_idx, layer in enumerate(layers):
                        if target_layer_flags[layer_idx]:
                            layer_parts = layer.split(".")
                            if (
                                layer_number_idx >= len(layer_parts)
                                or not layer_parts[layer_number_idx].isdigit()
                            ):
                                continue  # skip layers without number
                            base_name = ".".join(
                                layer_parts[:layer_number_idx]
                                + layer_parts[layer_number_idx + 1 :]
                            )
                            layer_number = int(layer_parts[layer_number_idx])
                            new_layer_number = random_shuffle_mapping[model][base_name][
                                layer_number
                            ]
                            new_layer_name = ".".join(
                                layer_parts[:layer_number_idx]
                                + [new_layer_number]
                                + layer_parts[layer_number_idx + 1 :]
                            )
                            temp_state_dicts[model][new_layer_name] = state_dicts[
                                model
                            ][layer]
                            layer_mappings[model][new_layer_name] = layer
                state_dicts = temp_state_dicts
            else:
                layer_numbers = {}
                for layer_idx, layer in enumerate(layers):
                    if target_layer_flags[layer_idx]:
                        layer_parts = layer.split(".")
                        for part in layer_parts:
                            if part.isdigit():
                                layer_numbers[layer] = int(part)
                                break  # Only consider the first number for each layer
                if layer_numbers:
                    max_layer_number = max(layer_numbers.values())
                else:
                    max_layer_number = 0
                temp_state_dicts = deepcopy(state_dicts)
                # Wrap around and shift each model by a different amount
                for model_idx, model in enumerate(models):
                    for layer_idx, layer in enumerate(layers):
                        target_flag = target_layer_flags[layer_idx]
                        if not target_flag:
                            continue
                        layer_number = layer_numbers.get(layer)
                        if layer_number is None:
                            continue
                        new_layer_number = (
                            layer_number + model_idx * self.config.shift_layers
                        ) % (max_layer_number + 1)
                        new_layer_parts = []
                        replaced = False  # Only replace the first numeric part FIXME: make it more robust
                        for part in layer.split("."):
                            if part.isdigit() and not replaced:
                                new_layer_parts.append(str(new_layer_number))
                                replaced = True
                            else:
                                new_layer_parts.append(part)
                        new_layer = ".".join(new_layer_parts)
                        temp_state_dicts[model][new_layer] = state_dicts[model][layer]
                state_dicts = temp_state_dicts

        if self.debug >= 2:
            # for evaluating pairwise cosine similarity
            unmerged_task_vectors = deepcopy(state_dicts)

        # compress
        for layer_idx, layer in enumerate(layers):
            shape = state_dicts[models[0]][layer].shape
            compressed_layer = None
            target_flag = target_layer_flags[layer_idx]
            # self.verbose = 1
            if self.verbose >= 1:
                log.info(f"{layer} | {shape} | {target_flag}")
            if not target_flag:
                if absorber is not None:
                    compressed_layer = absorber[layer]
                else:
                    for model in models:
                        if compressed_layer is None:
                            compressed_layer = deepcopy(state_dicts[model][layer])
                        else:
                            compressed_layer += deepcopy(state_dicts[model][layer])
            else:
                if self.mode == "random_binary_diagonal_matrix":
                    for model_idx, model in enumerate(models):
                        if self.different_across_layers:
                            seed = self.random_seed + model_idx + hash(layer) % 1e6
                        else:
                            seed = self.random_seed + model_idx
                        numpy_state = np.random.get_state()
                        np.random.seed(int(seed))
                        context = (
                            np.random.binomial(p=0.5, n=1, size=(1, shape[-1])).astype(
                                np.float32
                            )
                            * 2
                            - 1
                        )
                        context = torch.from_numpy(context)
                        np.random.set_state(numpy_state)
                        compression_context[model][
                            layer
                        ] = context  # for analysis purposes
                        retrieval_context[model][layer] = context
                        if compressed_layer is None:
                            compressed_layer = state_dicts[model][layer] * context
                        else:
                            compressed_layer += state_dicts[model][layer] * context
                        if self.debug >= 2:
                            # hadamard product is not linear, convert it back to diagonal matrix and apply matrix multiplication
                            context_diag = torch.diag(context.squeeze())
                            unmerged_task_vectors[model][layer] = (
                                unmerged_task_vectors[model][layer] @ context_diag
                            )
                elif self.mode == "random_rotation_matrix":
                    for model_idx, model in enumerate(models):
                        if self.different_across_layers:
                            seed = self.random_seed + model_idx + hash(layer) % 1e6
                        else:
                            seed = self.random_seed + model_idx
                        context = torch.from_numpy(
                            ortho_group.rvs(shape[-1], random_state=seed).astype(
                                "float32"
                            )
                        )
                        compression_context[model][
                            layer
                        ] = context  # for analysis purposes
                        retrieval_context[model][layer] = context.t()
                        if compressed_layer is None:
                            compressed_layer = state_dicts[model][layer] @ context
                        else:
                            compressed_layer += state_dicts[model][layer] @ context
                        if self.debug >= 2:
                            unmerged_task_vectors[model][layer] = (
                                unmerged_task_vectors[model][layer] @ context
                            )
                elif self.mode == "random_dense_matrix":
                    for model_idx, model in enumerate(models):
                        if self.different_across_layers:
                            seed = self.random_seed + model_idx + hash(layer) % 1e6
                        else:
                            seed = self.random_seed + model_idx
                        numpy_state = np.random.get_state()
                        np.random.seed(int(seed))
                        context = torch.from_numpy(
                            np.random.randn(shape[-1], shape[-1]).astype(np.float32)
                        )
                        np.random.set_state(numpy_state)
                        compression_context[model][
                            layer
                        ] = context  # for analysis purposes
                        retrieval_context[model][layer] = torch.linalg.pinv(
                            context.to("cuda")
                        ).to("cpu")
                        if compressed_layer is None:
                            compressed_layer = state_dicts[model][layer] @ context
                        else:
                            compressed_layer += state_dicts[model][layer] @ context
                        if self.debug >= 2:
                            unmerged_task_vectors[model][layer] = (
                                unmerged_task_vectors[model][layer] @ context
                            )
                elif self.mode == "random_diagonal_matrix":
                    for model_idx, model in enumerate(models):
                        if self.different_across_layers:
                            seed = self.random_seed + model_idx + hash(layer) % 1e6
                        else:
                            seed = self.random_seed + model_idx
                        numpy_state = np.random.get_state()
                        np.random.seed(int(seed))
                        context = torch.from_numpy(
                            np.random.randn(1, shape[-1]).astype(np.float32)
                        )
                        np.random.set_state(numpy_state)
                        compression_context[model][
                            layer
                        ] = context  # for analysis purposes
                        retrieval_context[model][layer] = 1 / context
                        if compressed_layer is None:
                            compressed_layer = state_dicts[model][layer] * context
                        else:
                            compressed_layer += state_dicts[model][layer] * context
                        if self.debug >= 2:
                            unmerged_task_vectors[model][layer] = (
                                unmerged_task_vectors[model][layer] * context
                            )
                elif self.mode == "identity_matrix":
                    for model_idx, model in enumerate(models):
                        context = torch.eye(shape[-1])
                        compression_context[model][
                            layer
                        ] = context  # for analysis purposes
                        retrieval_context[model][layer] = context
                        if compressed_layer is None:
                            compressed_layer = state_dicts[model][layer] @ context
                        else:
                            compressed_layer += state_dicts[model][layer] @ context
                        if self.debug >= 2:
                            unmerged_task_vectors[model][layer] = (
                                unmerged_task_vectors[model][layer] @ context
                            )
                else:
                    raise ValueError(f"Unsupported mode: {self.mode}")

            compressed_layers[layer] = compressed_layer

        # retrieve: for purpose of benchmarking, retrieve all models at once. In practice, retrieval should be done per model request.
        nonzero_param_count = 0
        nonzero_param_count_context = 0
        total_bytes_retrieved = 0

        if self.debug >= 2:
            for model in test_models:
                tv_new[model] = deepcopy(unmerged_task_vectors)

        for layer_idx, layer in enumerate(layers):
            shape = state_dicts[models[0]][layer].shape
            target_flag = target_layer_flags[layer_idx]
            if not target_flag:
                if mode == "superposed_model_soup":
                    # we don't count non-target layers for superposed task arithmetic
                    # because they can be absorbed into the pretrained weights
                    param_count = torch.numel(compressed_layers[layer])
                    total_bytes_retrieved += (
                        param_count * compressed_layers[layer].element_size()
                    )
                    nonzero_param_count += param_count
                for model in models:
                    retrieval_models[model][layer] = compressed_layers[layer]
            else:
                if (
                    mode == "superposed_task_arithmetic"
                    and self.mode == "identity_matrix"
                    and self.shift_layers == 0
                ):
                    # we don't count target layers for task arithmetic
                    # because they can be absorbed into the pretrained weights
                    pass
                else:
                    param_count = torch.numel(compressed_layers[layer])
                    total_bytes_retrieved += (
                        param_count * compressed_layers[layer].element_size()
                    )
                    nonzero_param_count += torch.numel(compressed_layers[layer])

                if self.mode in [
                    "random_binary_diagonal_matrix",
                    "random_rotation_matrix",
                    "random_dense_matrix",
                    "random_diagonal_matrix",
                    "identity_matrix",
                ]:
                    for model in models:
                        if self.mode not in ["identity_matrix"]:
                            nonzero_count = torch.numel(retrieval_context[model][layer])
                            if self.mode == "random_binary_diagonal_matrix":
                                total_bytes_retrieved += (
                                    nonzero_count * 1
                                )  # 1 byte per element for binary
                            else:
                                total_bytes_retrieved += (
                                    nonzero_count
                                    * retrieval_context[model][layer].element_size()
                                )
                            nonzero_param_count += nonzero_count
                            nonzero_param_count_context += nonzero_count
                        if retrieval_context[model][layer].shape[0] == 1:
                            retrieval_models[model][layer] = (
                                compressed_layers[layer]
                                * retrieval_context[model][layer]
                            )
                        else:
                            retrieval_models[model][layer] = (
                                compressed_layers[layer]
                                @ retrieval_context[model][layer]
                            )
                        if self.debug >= 2 and model in test_models:
                            if retrieval_context[model][layer].shape[0] == 1:
                                retrieval_context_diag = torch.diag(
                                    retrieval_context[model][layer].squeeze()
                                )
                                for m in models:
                                    tv_new[model][m][layer] = (
                                        tv_new[model][m][layer] @ retrieval_context_diag
                                    )
                            else:
                                for m in models:
                                    tv_new[model][m][layer] = (
                                        tv_new[model][m][layer]
                                        @ retrieval_context[model][layer]
                                    )
                else:
                    raise ValueError(f"Unsupported mode: {self.mode}")
        # for model in test_models:
        #     # print(retrieval_context[model]['vision_model.encoder.layers.4.self_attn.q_proj.weight'])
        # #     print('a')
        #     print(tv_new[model][models[3]]['vision_model.encoder.layers.4.self_attn.q_proj.weight'])

        # Shift the layers back
        if self.shift_layers != 0:
            if self.shift_layers == -1:  # random shuffling
                if self.debug >= 2:
                    temp_tv_new = deepcopy(tv_new)
                temp_retrieval_models = deepcopy(retrieval_models)
                for model_idx, model in enumerate(models):
                    # reverse_layer_mapping = {shuffled: original for original, shuffled in layer_mappings[model].items()}
                    for shuffled_layer, original_layer in layer_mappings[model].items():
                        temp_retrieval_models[model][original_layer] = retrieval_models[
                            model
                        ][shuffled_layer]
                        if self.debug >= 2 and model in test_models:
                            for m in models:
                                temp_tv_new[model][m][original_layer] = tv_new[model][
                                    m
                                ][shuffled_layer]
                retrieval_models = temp_retrieval_models
                if self.debug >= 2:
                    tv_new = temp_tv_new
            else:  # TODO: check the correctness of this mode
                # raise NotImplementedError("Shift back mode not implemented yet. No tv_new support yet.")
                if self.debug >= 2:
                    temp_tv_new = deepcopy(tv_new)
                temp_retrieval_models = deepcopy(retrieval_models)
                for model_idx, model in enumerate(models):
                    for layer_idx, layer in enumerate(layers):
                        target_flag = target_layer_flags[layer_idx]
                        if not target_flag:
                            continue
                        layer_parts = layer.split(".")
                        layer_number = None
                        for part in layer_parts:
                            if part.isdigit():
                                layer_number = int(part)
                                break  # Only consider the first number
                        if layer_number is None:
                            continue
                        new_layer_number = (
                            layer_number - model_idx * self.shift_layers
                        ) % (max_layer_number + 1)
                        new_layer_parts = []
                        replaced = False
                        for part in layer_parts:
                            if part.isdigit() and not replaced:
                                new_layer_parts.append(str(new_layer_number))
                                replaced = True  # Only replace the first numeric part
                            else:
                                new_layer_parts.append(part)
                        new_layer = ".".join(new_layer_parts)
                        temp_retrieval_models[model][new_layer] = retrieval_models[
                            model
                        ][layer]
                        if self.debug >= 2 and model in test_models:
                            for m in models:
                                temp_tv_new[model][m][new_layer] = tv_new[model][m][
                                    layer
                                ]
                retrieval_models = temp_retrieval_models
                if self.debug >= 2:
                    tv_new = temp_tv_new

        # for model in test_models:
        #     # print(retrieval_context[model]['vision_model.encoder.layers.4.self_attn.q_proj.weight'])
        # #     print('a')
        #     print(tv_new[model][models[3]]['vision_model.encoder.layers.4.self_attn.q_proj.weight'])

        # metadata
        if self.debug >= 2:
            if self.mode in [
                "random_binary_diagonal_matrix",
                "random_rotation_matrix",
                "random_dense_matrix",
                "random_diagonal_matrix",
                "identity_matrix",
            ]:
                layers = list(layers_old[models[0]].keys())
                layers_old_flattened = [
                    torch.cat([layers_old[model][layer].flatten() for layer in layers])
                    for model in models
                ]
                metadata["pairwise_cosine_similarity_matrix_before"] = (
                    pairwise_cosine_similarity_matrix(layers_old_flattened).tolist()
                )
                metadata["task_vector_dim"] = layers_old_flattened[0].shape[0]
                # layers_new = deepcopy(retrieval_models)
                rms = []
                for retrieval_model in test_models:
                    print(f"Retrieval model: {retrieval_model}")
                    layers_new_flattened = [
                        torch.cat(
                            [
                                tv_new[retrieval_model][model][layer].flatten()
                                for layer in layers
                            ]
                        )
                        for model in models
                    ]
                    rms.append(layers_new_flattened)
                    # print(layers_new_flattened[0][:50])
                    # layers_new_flattened = [torch.cat([layers_new[retrieval_model][layer].flatten() for layer in layers]) for model in models]
                    pcsm = pairwise_cosine_similarity_matrix(
                        layers_new_flattened
                    ).tolist()
                    print(pcsm)
                    metadata[
                        f"pairwise_cosine_similarity_matrix_after_{retrieval_model}"
                    ] = pcsm
        if self.debug >= 0:
            metadata["nonzero_parameter_count"] = (
                nonzero_param_count.item()
                if isinstance(nonzero_param_count, torch.Tensor)
                else nonzero_param_count
            )
            metadata["nonzero_param_count_context"] = (
                nonzero_param_count_context.item()
                if isinstance(nonzero_param_count_context, torch.Tensor)
                else nonzero_param_count_context
            )
            gbs = total_bytes_retrieved / 1e9
            metadata["total_gb_retrieved"] = gbs

        return retrieval_models, metadata



import logging
import os
from typing import Dict, List, Optional, cast

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
import torch
from numpy.typing import NDArray
from torch import nn
from tqdm.auto import tqdm

from fusion_bench import BaseAlgorithm, BaseModelPool, StateDictType, timeit_context
from fusion_bench.mixins import (
    LightningFabricMixin,
    SimpleProfilerMixin,
    auto_register_config,
)
from fusion_bench.utils import state_dict_to_vector, trainable_state_dict
from fusion_bench.utils.state_dict_arithmetic import state_dict_sub

log = logging.getLogger(__name__)


@auto_register_config
class TaskVectorViolinPlot(
    LightningFabricMixin,
    SimpleProfilerMixin,
    BaseAlgorithm,
):
    """
    Creates violin plots to visualize the distribution of task vector values across models.

    This class implements the task vector visualization technique described in:
    "Efficient and Effective Weight-Ensembling Mixture of Experts for Multi-Task Model Merging"
    by L. Shen, A. Tang, E. Yang et al. (https://arxiv.org/abs/2410.21804)

    Task vectors represent the parameter differences between fine-tuned models and their
    pretrained base model, computed as:
        task_vector = finetuned_params - pretrained_params

    The algorithm generates two types of violin plots:
    1. Distribution of raw task vector values (positive and negative)
    2. Distribution of absolute task vector values

    Args:
        trainable_only (bool): If True, only consider trainable parameters when computing
            task vectors. If False, use all parameters.
        max_points_per_model (int, optional): Maximum number of parameters to sample
            per model for memory efficiency. If None or 0, uses all parameters.
            Defaults to 1000.
        fig_kwargs (dict, optional): Dictionary of keyword arguments to pass to
            matplotlib.pyplot.subplots. Common options include:
            - figsize: Tuple of (width, height) in inches
            - dpi: Dots per inch for resolution
            - facecolor: Figure background color
            Defaults to None.
        output_path (str, optional): Directory to save the violin plots. If None,
            uses the fabric logger's log directory. Defaults to None.

    Outputs:
        - task_vector_violin.pdf: Violin plot of raw task vector value distributions
        - task_vector_violin_abs.pdf: Violin plot of absolute task vector value distributions

    Returns:
        The pretrained model from the model pool.

    Example:
        ```python
        plotter = TaskVectorViolinPlot(
            trainable_only=True,
            max_points_per_model=5000,
            fig_kwargs={'figsize': (12, 8), 'dpi': 300},
            output_path='./analysis_plots'
        )
        pretrained_model = plotter.run(modelpool)
        ```

    Note:
        This visualization is particularly useful for understanding:
        - How different tasks affect model parameters
        - The magnitude and distribution of parameter changes
        - Similarities and differences between task adaptations
    """

    # config_mapping is a mapping from the attributes to the key in the configuration files
    _config_mapping = BaseAlgorithm._config_mapping | {
        "_output_path": "output_path",
    }

    def __init__(
        self,
        trainable_only: bool,
        max_points_per_model: Optional[int] = 1000,
        fig_kwawrgs=None,
        output_path: Optional[str] = None,
        **kwargs,
    ):
        """
        Initialize the TaskVectorViolinPlot analyzer.

        Args:
            trainable_only (bool): Whether to consider only trainable parameters when
                computing task vectors. Set to True to focus on learnable parameters,
                False to include all parameters including frozen ones.
            max_points_per_model (int, optional): Maximum number of parameter values
                to sample per model for visualization. Useful for large models to
                manage memory usage and plot clarity. Set to None or 0 to use all
                parameters. Defaults to 1000.
            fig_kwargs (dict, optional): Keyword arguments passed to matplotlib's
                subplots function for plot customization. Examples:
                - {'figsize': (10, 6)} for plot dimensions
                - {'dpi': 300} for high resolution
                - {'facecolor': 'white'} for background color
                Defaults to None (uses matplotlib defaults).
            output_path (str, optional): Directory path where violin plots will be saved.
                If None, uses the fabric logger's log directory. The directory will be
                created if it doesn't exist. Defaults to None.
            **kwargs: Additional keyword arguments passed to parent classes.

        Note:
            The parameter name 'fig_kwawrgs' appears to be a typo for 'fig_kwargs'.
            This should be corrected in the parameter name for consistency.
        """
        super().__init__(**kwargs)
        self._output_path = output_path

    @property
    def output_path(self):
        if self._output_path is None:
            return self.fabric.logger.log_dir
        else:
            return self._output_path

    def run(self, modelpool: BaseModelPool):
        """
        Execute the task vector violin plot analysis and visualization.

        This method implements the core algorithm that:
        1. Loads the pretrained base model from the model pool
        2. Computes task vectors for each fine-tuned model (parameter differences)
        3. Creates two violin plots showing the distribution of task vector values:
           - Raw values plot: Shows positive and negative parameter changes
           - Absolute values plot: Shows magnitude of parameter changes
        4. Saves both plots as PDF files in the output directory

        The visualization technique follows the approach described in:
        "Efficient and Effective Weight-Ensembling Mixture of Experts for Multi-Task Model Merging"

        Args:
            modelpool (BaseModelPool): Pool containing both a pretrained model and
                fine-tuned models. Must have `has_pretrained=True`.

        Returns:
            nn.Module: The pretrained model loaded from the model pool.

        Raises:
            AssertionError: If the model pool doesn't contain a pretrained model.

        Side Effects:
            - Creates output directory if it doesn't exist
            - Saves 'task_vector_violin.pdf' (raw values distribution)
            - Saves 'task_vector_violin_abs.pdf' (absolute values distribution)
            - Prints progress information during task vector computation

        Example Output Files:
            - task_vector_violin.pdf: Shows how parameters change (+ and -)
            - task_vector_violin_abs.pdf: Shows magnitude of parameter changes
        """
        assert modelpool.has_pretrained
        pretrained_model = modelpool.load_pretrained_model()

        # Compute task vectors for each fine-tuned model
        with torch.no_grad(), timeit_context("Computing task vectors"):
            task_vectors: Dict[str, NDArray] = {}
            for name, finetuned_model in tqdm(
                modelpool.named_models(), total=len(modelpool)
            ):
                print(f"computing task vectors for {name}")
                task_vectors[name] = self.get_task_vector(
                    pretrained_model, finetuned_model
                )

        # === Create violin plot ===
        fig, ax = plt.subplots(
            1, 1, **self.fig_kwargs if self.fig_kwargs is not None else {}
        )
        fig = cast(plt.Figure, fig)
        ax = cast(plt.Axes, ax)

        # Prepare data for plotting
        data = [values for values in task_vectors.values()]
        labels = list(task_vectors.keys())

        # Create violin plot using seaborn
        with timeit_context("ploting"):
            sns.violinplot(data=data, ax=ax)

        # Customize plot
        ax.set_xticklabels(labels, rotation=45, ha="right")
        ax.set_ylabel("Task Vector Values")
        ax.set_title("Distribution of Task Vector Values")

        # Adjust layout to prevent label cutoff and save plot
        plt.tight_layout()
        os.makedirs(self.output_path, exist_ok=True)
        output_file = f"{self.output_path}/task_vector_violin.pdf"
        plt.savefig(output_file, bbox_inches="tight")
        plt.close(fig)

        # === Create violin plot (Abs values) ===
        fig, ax = plt.subplots(
            1, 1, **self.fig_kwargs if self.fig_kwargs is not None else {}
        )
        fig = cast(plt.Figure, fig)
        ax = cast(plt.Axes, ax)

        # Prepare data for plotting
        data = [np.abs(values) for values in task_vectors.values()]
        labels = list(task_vectors.keys())

        # Create violin plot using seaborn
        with timeit_context("ploting abs value plot"):
            sns.violinplot(data=data, ax=ax)

        # Customize plot
        ax.set_xticklabels(labels, rotation=45, ha="right")
        ax.set_ylabel("The Absolute Values")
        ax.set_title("Distribution of Task Vector Absolute Values")

        # Adjust layout to prevent label cutoff and save plot
        plt.tight_layout()
        os.makedirs(self.output_path, exist_ok=True)
        output_file = f"{self.output_path}/task_vector_violin_abs.pdf"
        plt.savefig(output_file, bbox_inches="tight")
        plt.close(fig)

        return pretrained_model

    def get_task_vector(self, pretrained_model, finetuned_model):
        """
        Compute the task vector representing parameter changes from pretraining to fine-tuning.

        The task vector quantifies how model parameters have changed during task-specific
        fine-tuning and is computed as:
            task_vector = finetuned_params - pretrained_params

        Args:
            pretrained_model (nn.Module): The base pretrained model
            finetuned_model (nn.Module): The fine-tuned model for a specific task

        Returns:
            np.ndarray: Flattened numpy array containing parameter differences.
                If max_points_per_model is set, the array may be randomly downsampled
                for memory efficiency and visualization clarity.

        Processing Steps:
            1. Extract state dictionaries from both models
            2. Compute parameter differences (subtraction)
            3. Flatten to 1D vector
            4. Convert to numpy array with float32 precision
            5. Optionally downsample if max_points_per_model is specified

        Note:
            - Uses only trainable parameters if trainable_only=True
            - Downsampling uses random sampling without replacement
            - Preserves the relative distribution of parameter changes
        """
        task_vector = state_dict_sub(
            self.get_state_dict(finetuned_model),
            self.get_state_dict(pretrained_model),
        )
        task_vector = state_dict_to_vector(task_vector)

        task_vector = task_vector.cpu().float().numpy()
        # downsample if necessary
        if (
            self.max_points_per_model is not None
            and self.max_points_per_model > 0
            and task_vector.shape[0] > self.max_points_per_model
        ):
            log.info(
                f"Downsampling task vectors to {self.max_points_per_model} points."
            )
            indices = np.random.choice(
                task_vector.shape[0], self.max_points_per_model, replace=False
            )
            task_vector = task_vector[indices].copy()

        return task_vector

    def get_state_dict(self, model: nn.Module):
        """
        Extract the state dictionary from a model based on parameter filtering settings.

        Args:
            model (nn.Module): The PyTorch model to extract parameters from

        Returns:
            Dict[str, torch.Tensor]: State dictionary containing model parameters.
                If trainable_only=True, returns only parameters with requires_grad=True.
                If trainable_only=False, returns all parameters including frozen ones.

        Note:
            This method respects the trainable_only configuration to focus analysis
            on either learnable parameters or the complete parameter set depending
            on the research question being addressed.
        """
        if self.trainable_only:
            return trainable_state_dict(model)
        else:
            return model.state_dict()



from typing import Dict, List, Literal, Union

import torch
import torch.nn.functional as F
from torch import Tensor, nn

from fusion_bench.models import ParameterDictModel
from fusion_bench.utils.type import StateDictType


def mask_sparsity(mask: Dict[str, Tensor]):
    total = 0
    non_zero = 0
    for name, m in mask.items():
        total += m.numel()
        non_zero += m.sum().item()
    return non_zero / total


def to_state_dict(
    state_dict_or_model: Union[StateDictType, nn.Module],
    ignore_untrained_params: bool = False,
    ignore_keys: List[str] = [],
    keep_vars: bool = False,
):
    """
    Convert a PyTorch model or state dictionary to a state dictionary, optionally ignoring untrained parameters, specified keys, and keeping differentiable.

    Args:
        state_dict_or_model: Either a PyTorch model (nn.Module) or a state dictionary (StateDictType).
        ignore_untrained: If True, ignore parameters that are not being trained (i.e., those with requires_grad=False).
        ignore_keys: A list of keys to ignore when converting to a state dictionary.
        keep_vars: If True, keeps the Variable wrappers around the tensor data; otherwise, the returned state dictionary will have tensors only.


    Returns:
        A state dictionary (StateDictType) containing the model's parameters.
    """
    if isinstance(state_dict_or_model, nn.Module):
        state_dict: StateDictType = state_dict_or_model.state_dict(keep_vars=True)
        if ignore_untrained_params:
            for name, param in state_dict.items():
                if not param.requires_grad:
                    state_dict.pop(name)
    else:
        state_dict = state_dict_or_model

    for key in ignore_keys:
        state_dict.pop(key, None)

    if not keep_vars:
        for name, param in state_dict.items():
            state_dict[name] = param.detach()
    return state_dict


def get_masked_state_dict(
    state_dict_or_model: Union[StateDictType, nn.Module], mask: Dict[str, Tensor]
):
    state_dict = to_state_dict(state_dict_or_model)
    masked_state_dict = {}
    for name, m in mask.items():
        masked_state_dict[name] = state_dict[name] * m
    return masked_state_dict


class MaskModel(ParameterDictModel):
    def __init__(
        self,
        state_dict_or_model: Union[StateDictType, nn.Module],
        ignore_keys: List[str] = [],
        ignore_untrained_params: bool = True,
        parameter_type: Literal["probs", "logits"] = None,
    ):
        # Convert the model or state dictionary to a state dictionary
        state_dict = to_state_dict(
            state_dict_or_model,
            ignore_untrained_params=ignore_untrained_params,
            ignore_keys=ignore_keys,
        )

        # Convert the tensor dictionary to a parameter dictionary
        parameters = {}
        for name, param in state_dict.items():
            parameters[name] = nn.Parameter(torch.zeros_like(param), requires_grad=True)

        super().__init__(parameters)
        self.parameter_type = parameter_type

    def _param_to_distribution(
        self,
        param: Tensor,
        mask_type: Literal["discrete", "continuous"],
        temperature: float = 0.5,
        **kwargs,
    ):
        if self.parameter_type == "probs":
            if mask_type == "discrete":
                return torch.distributions.Bernoulli(probs=param, **kwargs)
            elif mask_type == "continuous":
                return torch.distributions.RelaxedBernoulli(
                    probs=param, temperature=temperature, **kwargs
                )
        elif self.parameter_type == "logits":
            if mask_type == "discrete":
                return torch.distributions.Bernoulli(logits=param, **kwargs)
            elif mask_type == "continuous":
                return torch.distributions.RelaxedBernoulli(
                    logits=param, temperature=temperature, **kwargs
                )
        raise ValueError(f"Invalid parameter type: {self.parameter_type}")

    def get_distribution(
        self,
        mask_type: Literal["discrete", "continuous"],
        **kwargs,
    ):
        return {
            name: self._param_to_distribution(param, mask_type=mask_type, **kwargs)
            for name, param in self.named_parameters()
        }

    def sample_mask(
        self,
        mask_type: Literal["discrete", "continuous"] = "discrete",
        **kwargs,
    ):
        mask = {}
        for name, param in self.named_parameters():
            dist = self._param_to_distribution(param, mask_type, **kwargs)
            if mask_type == "discrete":
                mask[name] = dist.sample()
            elif mask_type == "continuous":
                mask[name] = dist.rsample()
            else:
                raise ValueError(f"Invalid mask type: {mask_type}")
        return mask

    def fill_(self, value: float):
        """
        Fills the parameters with a given value.

        Args:
            value (float): The value to fill the parameters with.
        """
        for param in self.parameters():
            param.data.fill_(value)

    def clamp_(self, min=None, max=None):
        for param in self.parameters():
            param.data = param.data.clamp(min, max)

    def sparsity(self):
        mask = self.sample_mask(mask_type="discrete")
        total = 0
        non_zero = 0
        for name, m in mask.items():
            total += m.numel()
            non_zero += m.sum().item()
        return non_zero / total



from typing import Dict

import numpy as np


def compute_backward_transfer(
    acc_Ti: Dict[str, float], acc_ii: Dict[str, float]
) -> float:
    R"""
    Compute the backward transfer (BWT) of a model on a set of tasks.

    Equation:
        $BWT = \frac{1}{n} \sum_{k=1}^{n} (acc_{T,i}[k] - acc_{i,i}[k])$

    Returns:
        float: The backward transfer of the model.
    """
    assert set(acc_ii.keys()) == set(acc_Ti.keys())
    bwt = 0
    for task_name in acc_ii:
        bwt += acc_Ti[task_name] - acc_ii[task_name]
    return bwt / len(acc_ii)



R"""
Copy from https://github.com/Lucky-Lance/Expert_Sparsity/tree/main/model

Original repo: https://github.com/Lucky-Lance/Expert_Sparsity

Reference:
    Not All Experts are Equal: Efficient Expert Pruning and Skipping for Mixture-of-Experts Large Language Models.
    ACL 2024.
    http://arxiv.org/abs/2402.14800
"""

from .wrapper import (
    DynamicSkippingMixtralSparseMoeBlockWrapper,
    PrunableMixtralSparseMoeBlockWrapper,
)



from typing import Dict, Union

import torch
from torch import Tensor, nn

from fusion_bench.utils.type import StateDictType


def param_random_drop_(param: Tensor, sparsity_level: float, rescale: bool):
    """
    Randomly drops elements in the given tensor based on the sparsity level.

    Args:
        param (Tensor): The tensor whose elements are to be randomly dropped.
        sparsity_level (float): The fraction of elements to drop (between 0 and 1).
        rescale (bool): If True, rescale the remaining elements to maintain the original sum.

    Returns:
        None
    """
    mask = torch.rand_like(param) > sparsity_level
    param.data = param.data * mask
    if rescale:
        param.data = param.data / (1 - sparsity_level)


def module_random_drop_(
    tv: Union[nn.Module, StateDictType], sparsity_level: float, rescale: bool
):
    """
    Applies random drop to all parameters in a module or state dictionary.

    Args:
        tv (Union[nn.Module, StateDictType]): The module or state dictionary whose parameters are to be randomly dropped.
        sparsity_level (float): The fraction of elements to drop (between 0 and 1).
        rescale (bool): If True, rescale the remaining elements to maintain the original sum.

    Returns:
        None
    """
    if isinstance(tv, nn.Module):
        for param in tv.parameters():
            param_random_drop_(param, sparsity_level, rescale)
    else:
        for param in tv.values():
            param_random_drop_(param, sparsity_level, rescale)


def trainable_state_dict(module: nn.Module):
    """
    Returns a state dictionary containing only the trainable parameters of the given module.

    Args:
        module (nn.Module): The module from which to extract the trainable parameters.

    Returns:
        dict: A dictionary where the keys are parameter names and the values are the corresponding trainable parameters.
    """
    return {
        name: param for name, param in module.named_parameters() if param.requires_grad
    }


def module_sub_(
    a: Union[nn.Module, StateDictType],
    b: Union[nn.Module, StateDictType],
    trainable_only: bool = True,
):
    """
    Subtracts the parameters of module b from module a in-place.

    Args:
        a (nn.Module): The module whose parameters will be subtracted from.
        b (nn.Module): The module whose parameters will be subtracted.

    Returns:
        nn.Module: The modified module a with updated parameters.
    """
    for (a_name, a_param), (b_name, b_param) in zip(
        a.named_parameters() if isinstance(a, nn.Module) else a.items(),
        b.named_parameters() if isinstance(b, nn.Module) else b.items(),
    ):
        assert a_name == b_name, "Mismatch in parameter names"
        if trainable_only and not a_param.requires_grad:
            continue
        a_param.data = a_param.data - b_param.data
    return a



classname_mapping = {
    "0": "T - shirt / top",
    "1": "Trouser",
    "2": "Pullover",
    "3": "Dress",
    "4": "Coat",
    "5": "Sandal",
    "6": "Shirt",
    "7": "Sneaker",
    "8": "Bag",
    "9": "Ankle boot",
}
classnames = [classname_mapping[str(i)] for i in range(10)]

templates = [
    lambda c: f"a photo of a {c}.",
    lambda c: f"a photo of the {c}.",
]



import os
from typing import Literal, cast

import pandas as pd
import torch
from omegaconf import DictConfig
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm
from transformers import CLIPVisionModel

from fusion_bench import BaseAlgorithm, BaseModelPool, auto_register_config
from fusion_bench.dataset import CLIPDataset
from fusion_bench.method import SmileUpscalingAlgorithm
from fusion_bench.mixins import LightningFabricMixin, SimpleProfilerMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.taskpool.clip_vision.taskpool import LayerWiseFeatureSaver
from fusion_bench.utils.devices import clear_cuda_cache


@auto_register_config
class LowRankApproximation(BaseAlgorithm):
    def __init__(self, rank: int, device: str = "cuda", **kwargs):
        """Low-rank approximation of fine-tuned updates."""
        super().__init__(**kwargs)

    def run(self, modelpool: BaseModelPool):
        # Implement low-rank approximation logic here
        base_model = modelpool.load_pretrained_model()

        models = {}
        for model_name in tqdm(modelpool.model_names, "processing models"):
            task_model = modelpool.load_model(model_name)
            for module_name, module in task_model.named_modules():
                if isinstance(module, nn.Linear):
                    w = cast(
                        nn.Linear, base_model.get_submodule(module_name)
                    ).weight.to(dtype=torch.float32, device=self.device, copy=True)
                    w_ft = module.weight.to(
                        dtype=torch.float32, device=self.device, copy=True
                    )

                    # Compute low-rank approximation
                    w_diff = w_ft - w
                    u, s, vh = torch.linalg.svd(w_diff)
                    v = vh.T

                    u = u[:, : self.rank]
                    s = s[: self.rank]
                    v = v[:, : self.rank]

                    low_rank_w_diff = torch.linalg.multi_dot((u, torch.diag(s), v.T))
                    low_rank_w = w + low_rank_w_diff

                    module.weight.data = low_rank_w.to(
                        dtype=module.weight.dtype,
                        device=module.weight.device,
                    )

            models[model_name] = task_model
        return models


@auto_register_config
class ErrorAccumulationAnalysisForCLIP(
    LightningFabricMixin,
    BaseAlgorithm,
):
    def __init__(
        self,
        gate_k: int,
        k: int,
        seed: int = 42,
        top_k: int = 1,
        dataset_kwargs: DictConfig = None,
        max_samples: int = 1024,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if dataset_kwargs is None:
            self.dataset_kwargs = DictConfig(
                {
                    "batch_size": 32,
                    "num_workers": 4,
                }
            )

    def run(self, modelpool: CLIPVisionModelPool):
        assert self.fabric.world_size == 1, "Distributed inference is not supported."
        # get the smile model
        smile_algorithm = SmileUpscalingAlgorithm(
            gate_k=self.gate_k, k=self.k, top_k=self.top_k, device=self.fabric.device
        )
        smile_model = smile_algorithm.run(modelpool)
        # get low-rank models
        low_rank_models = LowRankApproximation(rank=self.k).run(modelpool)

        results = {
            "model_name": [],
            "method": [],
            "layer_index": [],
            "approximation_error": [],
        }

        for model_name in modelpool.model_names:
            dataset = modelpool.load_test_dataset(model_name)
            processor = modelpool.load_processor()
            dataset = CLIPDataset(dataset, processor)
            dataloader = DataLoader(dataset, shuffle=True, **self.dataset_kwargs)
            dataloader = self.fabric.setup_dataloaders(dataloader)

            # finetuned_model
            finetuned_model = modelpool.load_model(model_name)
            finetuned_model = self.to_device(finetuned_model)
            self.collect_hidden_states(
                finetuned_model,
                dataloader=dataloader,
                model_name=f"{model_name}/finetuned",
            )
            del finetuned_model
            clear_cuda_cache()

            # smile model
            smile_model = self.to_device(smile_model)
            self.collect_hidden_states(
                smile_model, dataloader=dataloader, model_name=f"{model_name}/smile"
            )
            smile_model.cpu()
            clear_cuda_cache()

            # low-rank models
            model = low_rank_models.pop(model_name)
            model = self.to_device(model)
            self.collect_hidden_states(
                model, dataloader=dataloader, model_name=f"{model_name}/low-rank"
            )
            del model
            clear_cuda_cache()

            del dataloader
            clear_cuda_cache()

    @torch.no_grad()
    def collect_hidden_states(
        self, model: CLIPVisionModel, dataloader, model_name: str
    ):
        self.fabric.seed_everything(
            self.seed, workers=True
        )  # make sure to get same data samples
        # register hooks
        hooks = {}
        hook_handles = {}
        for i, layer in enumerate(model.vision_model.encoder.layers):
            hooks[i] = LayerWiseFeatureSaver(
                save_path=os.path.join(self.log_dir, model_name, f"layer_{i}.pth"),
                first_token_only=True,
            )
            hook_handles[i] = layer.register_forward_hook(hooks[i])

        # forward pass
        num_total_samples = 0
        for images, _ in tqdm(dataloader, desc=f"Collecting features for {model_name}"):
            batch_size = images.size(0)
            model(images)
            num_total_samples += batch_size
            if num_total_samples >= self.max_samples:
                break

        # save features
        for i, hook in hooks.items():
            hook.save_features()

        # remove hooks
        for i, hook_handle in hook_handles.items():
            hook_handle.remove()

        return hooks



"""
This file contains the implementation of the regmean method for model merging.
modified from https://github.com/yule-BUAA/MergeLM/blob/6d49ad96fd69c92013654b837041b868aa806564/model_merging_methods/merging_methods.py
"""

import logging
import re
from collections import defaultdict
from typing import Dict, List, cast

import torch
from torch import Tensor, nn
from tqdm.autonotebook import tqdm

from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import SimpleProfilerMixin, auto_register_config
from fusion_bench.modelpool import BaseModelPool

from .utils import get_modules_to_merge, get_param_names_to_merge

log = logging.getLogger(__name__)


def reduce_non_diagonal_elements(
    regmean_weights: torch.Tensor, reduce_non_diagonal_ratio: float
):
    """
    reduce the non-diagonal elements in regmean_weights
    :param regmean_weights: Tensor, shape (hidden_dim, hidden_dim), input regmean weights
    :param reduce_non_diagonal_ratio: float, reduce non-diagonal elements in regmean weights by multiplying this scalar
    :return:
    """
    # diagonal matrix with (1 - reduce_non_diagonal_ratio) as elements
    diag_weights = torch.diag(
        torch.ones(regmean_weights.shape[0]) - reduce_non_diagonal_ratio
    ).to(regmean_weights.device)
    # matrix with reduce_non_diagonal_ratio as elements
    non_diag_weights = torch.zeros_like(diag_weights).fill_(reduce_non_diagonal_ratio)
    # diagonal elements are unchanged, while non-diagonal elements are multiplied by reduce_non_diagonal_ratio
    return regmean_weights * (diag_weights + non_diag_weights)


def merging_with_regmean_weights(
    models_to_merge_param_dict: dict,
    models_to_merge_regmean_weights_list: list,
    reduce_non_diagonal_ratio: float = 1.0,
    weight_transpose: bool = True,
):
    """
    merge parameters of different models with computed regmean weights

    Args:
        models_to_merge_param_dict: dict, dictionary of list, where key is the parameter name,
            value is a list of the corresponding parameters of all the models that need to be merged
        models_to_merge_regmean_weights_list: list, list of dictionaries with length len(models_to_merge),
            each dictionary records the regmean weights (matrix) of parameters for each model that needs to be merged, key is module name
        reduce_non_diagonal_ratio: float, reduce non-diagonal elements in regmean weights by multiplying this scalar

    Returns:
        dict: merged model parameters
    """
    # dict, dictionary of model parameters
    merged_params = {}

    for param_name, param_value_list in models_to_merge_param_dict.items():
        merged_by_regmean = False
        # only perform regmean merging on the "weight" parameter of Linear module
        if param_name.endswith(".weight"):
            module_name = param_name[: -len(".weight")]
            if module_name in models_to_merge_regmean_weights_list[0].keys():
                # two lists with length num_models_to_merge
                param_multiplied_results, module_regmean_weights_list = [], []
                for model_idx, model_to_merge_regmean_weights in enumerate(
                    models_to_merge_regmean_weights_list
                ):
                    # Tensor, shape (hidden_dim, hidden_dim)
                    module_regmean_weights = model_to_merge_regmean_weights[module_name]

                    # reduce non-diagonal elements
                    module_regmean_weights = reduce_non_diagonal_elements(
                        regmean_weights=module_regmean_weights,
                        reduce_non_diagonal_ratio=reduce_non_diagonal_ratio,
                    )
                    module_regmean_weights_list.append(module_regmean_weights)

                    model_to_merge_param = param_value_list[model_idx]
                    # since the weight shape of Linear module is (output_size, input_size), we need to transpose it
                    param_multiplied_results.append(
                        torch.matmul(
                            module_regmean_weights,
                            (
                                model_to_merge_param.transpose(0, 1)
                                if weight_transpose
                                else model_to_merge_param
                            ),
                        )
                    )

                # sum up module_regmean_weights and param_multiplied_results over all individual models
                sum_module_regmean_weights = sum(module_regmean_weights_list)
                sum_param_multiplied_results = sum(param_multiplied_results)

                # get the inverse matrix
                inv_sum_module_regmean_weights = torch.inverse(
                    sum_module_regmean_weights
                )
                # merge parameters with regmean
                merged_param = torch.matmul(
                    inv_sum_module_regmean_weights, sum_param_multiplied_results
                )
                # transpose to the original shape of "weight" in Linear module
                merged_params[param_name] = (
                    merged_param.transpose(0, 1) if weight_transpose else merged_param
                )
                merged_by_regmean = True
        # use average merging for parameters whose names are not end with ".weight" or not in Linear module
        if not merged_by_regmean:
            merged_params[param_name] = torch.stack(param_value_list, dim=0).mean(dim=0)

    return merged_params


def regmean_merging(
    models_to_merge: list,
    trainers: list,
    exclude_param_names_regex: list,
    nums_regmean_examples: list,
    reduce_non_diagonal_ratio: float = 1.0,
):
    """
    regmean merging method.

    Args:
        models_to_merge: list, individual models that need to be merged
        trainers: list, trainers of individual models
        exclude_param_names_regex: list, regular expression of names of parameters that need to be excluded
        nums_regmean_examples: list, numbers of examples to compute regmean weights
        reduce_non_diagonal_ratio: float, reduce non-diagonal elements in regmean weights by multiplying this scalar

    Returns:
        dict: merged model parameters
    """

    def compute_regmean_weights(module_name: str):
        """
        compute the regmean weights, a hook function to deal with each module's input
        :param module_name: str, module name
        :return:
        """

        def hook(module: nn.Module, input: tuple, output: torch.Tensor):
            # Tensor, shape (batch_size, sequence_length, hidden_dim)
            x = input[0].detach()
            batch_num_actual_examples = x.shape[0]
            # Tensor, shape (batch_size * sequence_length, hidden_dim)
            x = x.reshape(-1, x.shape[-1])
            # Tensor, shape (hidden_dim, hidden_dim)
            xtx = torch.matmul(x.transpose(0, 1), x)
            # store the averaged weights in regmean_weights
            if module_name not in regmean_weights.keys():
                regmean_weights[module_name] = xtx / x.shape[0]
                num_computed_examples[module_name] = x.shape[0]
                num_actual_examples[module_name] = batch_num_actual_examples
            else:
                regmean_weights[module_name] = (
                    regmean_weights[module_name] * num_computed_examples[module_name]
                    + xtx
                ) / (num_computed_examples[module_name] + x.shape[0])
                num_computed_examples[module_name] += x.shape[0]
                num_actual_examples[module_name] += batch_num_actual_examples

        return hook

    # dictionary of list, where key is the parameter name,
    # value is a list of the corresponding parameters of all the models that need to be merged
    models_to_merge_param_dict = defaultdict(list)

    # list of dictionaries with length len(models_to_merge),
    # each dictionary records the regmean weights (matrix) of parameters for each model that needs to be merged
    models_to_merge_regmean_weights_list = []

    # iterate each individual model that needs to be merged
    with torch.no_grad():
        for model_idx, (model_to_merge, trainer, num_regmean_examples) in enumerate(
            zip(models_to_merge, trainers, nums_regmean_examples)
        ):
            param_dict = {
                param_name: param_value
                for param_name, param_value in model_to_merge.named_parameters()
            }
            # exclude parameter whose name matches element in exclude_param_names_regex
            param_names_to_merge = get_param_names_to_merge(
                input_param_names=list(param_dict.keys()),
                exclude_param_names_regex=exclude_param_names_regex,
            )

            for param_name in param_names_to_merge:
                models_to_merge_param_dict[param_name].append(param_dict[param_name])

            linear_modules_to_merge = get_modules_to_merge(
                model=model_to_merge, include_module_types=[nn.Linear]
            )
            handles = []
            # dictionary, regmean matrices for each linear module inputs
            regmean_weights = {}
            # dictionary, number of examples (multiplied the sequence length) used for computing regmean matrices
            num_computed_examples = {}
            # dictionary, number of actual examples used for computing regmean matrices
            num_actual_examples = {}

            for module_name, linear_module_to_merge in linear_modules_to_merge.items():
                # register a hook in the forward process
                handle = linear_module_to_merge.register_forward_hook(
                    compute_regmean_weights(module_name=module_name)
                )
                handles.append(handle)

            train_dataloader = trainer.get_train_dataloader()
            if num_regmean_examples % trainer._train_batch_size != 0:
                print(
                    f"warning: the number of examples for computing regmean cannot be fully divided by the batch size for model {model_idx}, "
                    "which may lead to a slightly different number of the actually used examples."
                )
            for step, inputs in tqdm(
                enumerate(train_dataloader),
                desc=f"computing regmean weights for model {model_idx}",
            ):
                if (
                    len(num_actual_examples) > 0
                    and list(num_actual_examples.values())[0] >= num_regmean_examples
                ):
                    break
                inputs = trainer._prepare_inputs(inputs)
                outputs = model_to_merge(**inputs)

            models_to_merge_regmean_weights_list.append(regmean_weights)

            # remove the added hook
            for handle in handles:
                handle.remove()
        # merging with regmean weights
        merged_params = merging_with_regmean_weights(
            models_to_merge_param_dict=models_to_merge_param_dict,
            models_to_merge_regmean_weights_list=models_to_merge_regmean_weights_list,
            reduce_non_diagonal_ratio=reduce_non_diagonal_ratio,
        )

    return merged_params


@auto_register_config
class RegMeanAlgorithm(
    SimpleProfilerMixin,
    BaseAlgorithm,
):
    _include_module_type = [nn.Linear]

    def __init__(
        self,
        *,
        num_regmean_examples: int,
        exclude_param_names_regex: list,
        reduce_non_diagonal_ratio: float,
        weight_transpose: bool,
        **kwargs,
    ):
        super().__init__(**kwargs)

    def run(self, modelpool: BaseModelPool, **kwargs):
        if not isinstance(modelpool, BaseModelPool):
            modelpool = BaseModelPool(modelpool)
        self.modelpool = modelpool
        self.on_regmean_start()

        # dictionary of list, where key is the parameter name,
        # value is a list of the corresponding parameters of all the models that need to be merged
        models_to_merge_param_dict = defaultdict(list)

        # list of dictionaries with length len(models_to_merge),
        # each dictionary records the regmean weights (matrix) of parameters for each model that needs to be merged
        models_to_merge_regmean_weights_list = []

        param_names_to_merge = None

        with torch.no_grad():
            for name, model in modelpool.named_models():
                param_dict = model.state_dict()

                # exclude parameter whose name matches element in exclude_param_names_regex
                if param_names_to_merge is None:
                    param_names_to_merge = get_param_names_to_merge(
                        input_param_names=list(param_dict.keys()),
                        exclude_param_names_regex=self.config.get(
                            "exclude_param_names_regex", []
                        ),
                    )

                for param_name in param_names_to_merge:
                    models_to_merge_param_dict[param_name].append(
                        param_dict[param_name]
                    )

                linear_modules_to_merge = get_modules_to_merge(
                    model=model, include_module_types=self._include_module_type
                )
                assert len(linear_modules_to_merge) > 0, "No linear modules to merge"

                with (
                    self.profile("merging models"),
                    self.profile("computing regmean weights"),
                ):
                    regmean_weights = self.get_regmean_weights(
                        name,
                        model,
                        train_dataset=modelpool.load_train_dataset(name),
                        linear_modules_to_merge=linear_modules_to_merge,
                    )
                    models_to_merge_regmean_weights_list.append(regmean_weights)

        with self.profile("merging models"):
            # merging with regmean weights
            merged_params = merging_with_regmean_weights(
                models_to_merge_param_dict=models_to_merge_param_dict,
                models_to_merge_regmean_weights_list=models_to_merge_regmean_weights_list,
                reduce_non_diagonal_ratio=self.reduce_non_diagonal_ratio,
                weight_transpose=self.config.get("weight_transpose", True),
            )

            merged_model = modelpool.load_model("_pretrained_")
            merged_model.load_state_dict(merged_params, strict=False)

        self.print_profile_summary()
        return merged_model

    def on_regmean_start(self):
        pass

    def get_regmean_weights(
        self,
        model_name: str,
        model: nn.Module,
        train_dataset,
        linear_modules_to_merge: Dict[str, nn.Module],
    ):
        raise NotImplementedError



import functools
import logging
import os
from copy import deepcopy
from typing import Any, Iterator

import torch
from torch import Tensor
from torch.utils.data import DataLoader
from transformers import CLIPModel, CLIPProcessor
from transformers.models.clip.modeling_clip import CLIPEncoder

from fusion_bench.dataset import CLIPDataset
from fusion_bench.method.task_arithmetic.task_arithmetic import task_arithmetic_merge
from fusion_bench.mixins import CLIPClassificationMixin
from fusion_bench.modelpool import CLIPVisionModelPool
from fusion_bench.models.hf_clip import HFCLIPClassifier
from fusion_bench.models.we_moe import WeightEnsemblingMoE
from fusion_bench.tasks.clip_classification import get_classnames_and_templates
from fusion_bench.utils import timeit_context
from fusion_bench.utils.data import InfiniteDataLoader

from .we_moe import WeightEnsemblingMoEAlgorithm

log = logging.getLogger(__name__)


class CLIPWeightEnsemblingMoEAlgorithm(
    WeightEnsemblingMoEAlgorithm,
    CLIPClassificationMixin,
):
    """
    CLIPWeightEnsemblingMoEAlgorithm is a class that implements the WeightEnsemblingMoEAlgorithm
    for CLIP models. It extends the WeightEnsemblingMoEAlgorithm and CLIPClassificationMixin classes.

    Attributes:
        modelpool (CLIPVisionModelPool): The model pool containing the CLIP models.
    """

    modelpool: CLIPVisionModelPool = None

    def load_checkpoint(self, model: Any, checkpoint: Any):
        """
        Load the checkpoint file.

        Args:
            model: The model to load the checkpoint into.
            checkpoint: The path to the checkpoint file.
        """
        state = {"model": model}
        self._fabric.load(checkpoint, state)

    def save_checkpoint(self, model: Any, checkpoint: Any):
        """
        Save the checkpoint file.

        Args:
            model: The model to save the checkpoint from.
            checkpoint: The path to the checkpoint file.
        """
        self._fabric.save(checkpoint, {"model": model})

    def construct_moe_model(self) -> WeightEnsemblingMoE:
        """
        Construct the Mixture of Experts (MoE) model using the models in the model pool.

        Returns:
            WeightEnsemblingMoE: The constructed MoE model.
        """
        base_model = self.modelpool.load_model("_pretrained_")
        expert_models = [
            self.modelpool.load_model(m) for m in self.modelpool.model_names
        ]

        # Merge the models using task arithmetic
        moe_model = task_arithmetic_merge(
            # This function modifies the model in place, so we need to pass a deepcopy
            deepcopy(base_model),
            expert_models,
            scaling_factor=self.config.init_lambda,
        ).requires_grad_(False)

        # Up-scale MLP modules
        base_encoder: CLIPEncoder = base_model.vision_model.encoder
        moe_encoder: CLIPEncoder = moe_model.vision_model.encoder
        expert_encoders = [m.vision_model.encoder for m in expert_models]

        num_layers = len(base_encoder.layers)
        for layer_idx in range(num_layers):
            base_mlp = base_encoder.layers[layer_idx].mlp
            expert_mlps = [e.layers[layer_idx].mlp for e in expert_encoders]

            moe_encoder.layers[layer_idx].mlp = WeightEnsemblingMoE(
                hidden_size=base_encoder.config.hidden_size,
                base_model=base_mlp,
                expert_models=expert_mlps,
                init_lambda=self.config.init_lambda,
                batch_first=True,  # For open_clip models this is False
                router_hidden_layers=self.config.router_hidden_layers,
                batch_reduce=self.config.batch_reduce,
            )

        return moe_model

    @functools.cache
    def get_shuffled_test_loader_iter(self, tta_dataset: str) -> Iterator:
        """
        Get an iterator for the shuffled test data loader.

        Args:
            tta_dataset (str): The name of the test-time adaptation dataset.

        Returns:
            Iterator: An iterator for the shuffled test data loader.
        """
        dataset = self.modelpool.load_test_dataset(tta_dataset)
        dataset = CLIPDataset(dataset, processor=self.clip_processor)
        log.info("get_shuffled_test_loader_iter")
        loader = DataLoader(
            dataset,
            batch_size=self.config.batch_size,
            shuffle=True,
            num_workers=self.config.num_workers,
            pin_memory=True,
        )
        loader = self.fabric.setup_dataloaders(loader)
        return iter(InfiniteDataLoader(loader))

    def on_test_time_adaptation_start(self):
        """
        Load the CLIP processor and construct the zero-shot classification head for each task.
        """
        self.setup_zero_shot_classification_head()

    def compute_logits(self, module: Any, batch: Any, task: Any) -> Tensor:
        """
        Compute the logits for the given batch and task.

        Args:
            module: The model module.
            batch: The input batch.
            task: The task name.

        Returns:
            Tensor: The computed logits.
        """
        images, _ = batch
        text_embeds = self.zeroshot_weights[task]

        image_embeds = module(images)[1]
        image_embeds = self.visual_projection(image_embeds)

        # Normalize embeddings
        image_embeds = image_embeds / image_embeds.norm(p=2, dim=-1, keepdim=True)

        # Cosine similarity
        logits_per_text = (
            torch.matmul(text_embeds, image_embeds.t()) * self.logit_scale_exp
        )
        logits_per_image = logits_per_text.t()

        return logits_per_image



import json
import logging
import os
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Dict,
    Generic,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import torch
from accelerate import init_empty_weights
from accelerate.utils.constants import SAFE_WEIGHTS_NAME, WEIGHTS_NAME
from huggingface_hub import snapshot_download
from safetensors import safe_open
from safetensors.torch import load_file
from torch import nn
from torch.nn.modules.module import _IncompatibleKeys
from transformers import AutoConfig

from fusion_bench.utils.dtype import parse_dtype
from fusion_bench.utils.packages import import_object
from fusion_bench.utils.type import TorchModelType

if TYPE_CHECKING:
    from transformers import PretrainedConfig

log = logging.getLogger(__name__)

__all__ = ["resolve_checkpoint_path", "LazyStateDict"]


def resolve_checkpoint_path(
    checkpoint: str,
    hf_revision: Optional[str] = None,
    hf_cache_dir: Optional[str] = None,
    hf_proxies: Optional[Dict] = None,
):
    # If it's a local file or directory, return as is
    if os.path.exists(checkpoint):
        return checkpoint
    # If it's a HuggingFace Hub model id, download snapshot
    try:
        # This will download the model to the cache and return the local path
        local_path = snapshot_download(
            repo_id=checkpoint,
            revision=hf_revision,
            cache_dir=hf_cache_dir,
            proxies=hf_proxies,
        )
        return local_path
    except Exception as e:
        raise FileNotFoundError(
            f"Could not resolve checkpoint: {checkpoint}. Error: {e}"
        )


class LazyStateDict(Mapping[str, torch.Tensor], Generic[TorchModelType]):
    """
    Dictionary-like object that lazily loads a state dict from a checkpoint path.
    """

    _local_path: str
    """local path to the checkpoint."""
    _state_dict_cache: Optional[Dict]
    """Cache for the state dict, if enabled."""
    _index_filename: Optional[str]
    _checkpoint_files: Optional[List[str]]
    _index: Optional[Dict[str, str]]
    """Mapping of parameter names to checkpoint files."""

    def __init__(
        self,
        checkpoint: str,
        meta_module_class: Optional[Type[TorchModelType]] = None,
        meta_module: Optional[TorchModelType] = None,
        cache_state_dict: bool = False,
        torch_dtype: Optional[torch.dtype] = None,
        device: str = "cpu",
        hf_revision: Optional[str] = None,
        hf_cache_dir: Optional[str] = None,
        hf_proxies: Optional[Dict] = None,
    ):
        """
        Args:
            checkpoint (str): Path to the checkpoint file or directory.
            meta_module_class (Type[nn.Module], optional): Class of the meta module to instantiate.
            meta_module (nn.Module, optional): Pre-initialized meta module.
            cache_state_dict (bool): Whether to cache the state dict in memory.
            torch_dtype (torch.dtype, optional): The dtype to use for the tensors.
            device (str): The device to load the tensors onto.
            hf_revision (str, optional): The revision of the model to download from Hugging Face Hub.
            hf_cache_dir (str, optional): The cache directory for Hugging Face models.
            hf_proxies (Dict, optional): Proxies to use for downloading from Hugging Face Hub.
        """
        self.cache_state_dict = cache_state_dict

        # Validate that both meta_module_class and meta_module are not provided
        if meta_module_class is not None and meta_module is not None:
            raise ValueError(
                "Cannot provide both meta_module_class and meta_module, please provide only one."
            )

        self.meta_module_class = meta_module_class
        if isinstance(self.meta_module_class, str):
            self.meta_module_class = import_object(self.meta_module_class)
        self.meta_module = meta_module

        if self.meta_module_class is not None:
            with init_empty_weights():
                self.meta_module = self.meta_module_class.from_pretrained(
                    checkpoint,
                    torch_dtype=torch_dtype,
                    revision=hf_revision,
                    cache_dir=hf_cache_dir,
                    proxies=hf_proxies,
                )

        self._checkpoint = checkpoint
        self._local_path = resolve_checkpoint_path(
            checkpoint,
            hf_revision=hf_revision,
            hf_cache_dir=hf_cache_dir,
            hf_proxies=hf_proxies,
        )

        self._index, self._index_filename, self._checkpoint_files = (
            self._resolve_checkpoint_files(self._local_path)
        )

        if self._index is not None:
            # if meta_module is provided, remove the keys that are not in the meta_module
            if self.meta_module is not None:
                meta_module_state_dict = self.meta_module.state_dict()
                for key in tuple(self._index.keys()):
                    if key not in meta_module_state_dict:
                        self._index.pop(key)
            if cache_state_dict:
                self._state_dict_cache = {}
            else:
                self._state_dict_cache = None
        elif len(self._checkpoint_files) == 1 and self._checkpoint_files[0].endswith(
            SAFE_WEIGHTS_NAME
        ):
            # let the keys of self._index be the keys of the state dict, the values are the checkpoint file
            with safe_open(
                self._checkpoint_files[0], framework="pt", device=device
            ) as f:
                self._index = {key: self._checkpoint_files[0] for key in f.keys()}
                if cache_state_dict:
                    self._state_dict_cache = {}
                else:
                    self._state_dict_cache = None
        elif len(self._checkpoint_files) == 1 and self._checkpoint_files[0].endswith(
            WEIGHTS_NAME
        ):
            log.info(f"Loading full state dict from {WEIGHTS_NAME}")
            self._state_dict_cache = torch.load(self._checkpoint_files[0])
            # if meta_module is provided, remove the keys that are not in the meta_module
            if self.meta_module is not None:
                meta_module_state_dict = self.meta_module.state_dict()
                for key in tuple(self._state_dict_cache.keys()):
                    if key not in meta_module_state_dict:
                        self._state_dict_cache.pop(key)
        else:
            raise ValueError(
                f"Cannot determine the type of checkpoint, please provide a checkpoint path to a file containing a whole state dict with file name {WEIGHTS_NAME} or {SAFE_WEIGHTS_NAME}, or the index of a sharded checkpoint ending with `.index.json`."
            )

        self._torch_dtype = parse_dtype(torch_dtype)
        self._device = device

    @property
    def checkpoint(self) -> str:
        return self._checkpoint

    @property
    def config(self) -> "PretrainedConfig":
        return AutoConfig.from_pretrained(self._checkpoint)

    @property
    def dtype(self) -> torch.dtype:
        """
        `torch.dtype`: The dtype of the module (assuming that all the module parameters have the same dtype).
        """
        if hasattr(self, "_cached_dtype"):
            return self._cached_dtype

        first_key = next(iter(self.keys()))
        first_param = self[first_key]
        self._cached_dtype = first_param.dtype
        return self._cached_dtype

    def state_dict(self, keep_vars: bool = False) -> "LazyStateDict":
        """
        Args:
            keep_vars (bool): Ignored, as LazyStateDict does not support keep_vars. Just for compatibility.
        """
        return deepcopy(self)

    def _resolve_checkpoint_files(self, checkpoint: str):
        # reference: https://huggingface.co/docs/accelerate/v0.17.1/en/usage_guides/big_modeling
        checkpoint_files = None
        index_filename = None
        if os.path.isfile(checkpoint):
            if str(checkpoint).endswith(".json"):
                index_filename = checkpoint
            else:
                checkpoint_files = [checkpoint]
        elif os.path.isdir(checkpoint):
            # check if the whole state dict is present
            potential_state_bin = [
                f for f in os.listdir(checkpoint) if f == WEIGHTS_NAME
            ]
            potential_state_safetensor = [
                f for f in os.listdir(checkpoint) if f == SAFE_WEIGHTS_NAME
            ]
            if len(potential_state_bin) == 1:
                checkpoint_files = [os.path.join(checkpoint, potential_state_bin[0])]
            elif len(potential_state_safetensor) == 1:
                checkpoint_files = [
                    os.path.join(checkpoint, potential_state_safetensor[0])
                ]
            else:
                # otherwise check for sharded checkpoints
                potential_index = [
                    f for f in os.listdir(checkpoint) if f.endswith(".index.json")
                ]
                if len(potential_index) == 0:
                    raise ValueError(
                        f"{checkpoint} is not a folder containing a `.index.json` file or a {WEIGHTS_NAME} or a {SAFE_WEIGHTS_NAME} file"
                    )
                elif len(potential_index) == 1:
                    index_filename = os.path.join(checkpoint, potential_index[0])
                else:
                    raise ValueError(
                        f"{checkpoint} containing more than one `.index.json` file, delete the irrelevant ones."
                    )
        else:
            raise ValueError(
                "`checkpoint` should be the path to a file containing a whole state dict, or the index of a sharded "
                f"checkpoint, or a folder containing a sharded checkpoint or the whole state dict, but got {checkpoint}."
            )

        if index_filename is not None:
            checkpoint_folder = os.path.split(index_filename)[0]
            with open(index_filename) as f:
                index = json.loads(f.read())

            if "weight_map" in index:
                index = index["weight_map"]
            checkpoint_files = sorted(list(set(index.values())))
            checkpoint_files = [
                os.path.join(checkpoint_folder, f) for f in checkpoint_files
            ]
        else:
            index = None
        return index, index_filename, checkpoint_files

    def _load_tensor_from_checkpoint_file(
        self, checkpoint_file: str, key: str, update_cache: bool = True
    ) -> torch.Tensor:
        if checkpoint_file.endswith(".safetensors"):
            with safe_open(checkpoint_file, framework="pt", device=self._device) as f:
                tensor = f.get_tensor(key)
                if self._torch_dtype is not None:
                    tensor = tensor.to(self._torch_dtype)
                if update_cache and self._state_dict_cache is not None:
                    self._state_dict_cache[key] = tensor
                return tensor
        else:
            state_dict = torch.load(checkpoint_file, map_location=self._device)
            if update_cache:
                if self._state_dict_cache is not None:
                    self._state_dict_cache.update(state_dict)
                else:
                    log.warning(
                        f"Load full state dict from file {checkpoint_file}, but state dict cache is disabled."
                    )
            return state_dict[key]

    def __getitem__(self, key: str) -> torch.Tensor:
        if self._state_dict_cache is not None and key in self._state_dict_cache:
            return self._state_dict_cache[key]

        if self._index is None:
            if len(self._checkpoint_files) == 1 and os.path.isfile(
                self._checkpoint_files[0]
            ):
                checkpoint_file = self._checkpoint_files[0]
                tensor = self._load_tensor_from_checkpoint_file(
                    checkpoint_file, key, update_cache=True
                )
                return tensor
            else:
                if len(self._checkpoint_files) > 1:
                    raise RuntimeError(
                        "Get multiple checkpoint files, but index is not provided."
                    )
                if not os.path.isfile(self._checkpoint_files[0]):
                    raise FileNotFoundError(
                        f"Checkpoint file {self._checkpoint_files[0]} not found."
                    )
                raise RuntimeError("Unexpected error.")
        else:
            if key not in self._index:
                raise KeyError(f"Key {key} not found in index.")
            checkpoint_file = os.path.join(self._local_path, self._index[key])
            if not os.path.isfile(checkpoint_file):
                raise FileNotFoundError(f"Checkpoint file {checkpoint_file} not found.")
            tensor = self._load_tensor_from_checkpoint_file(
                checkpoint_file, key, update_cache=True
            )
            return tensor

    def pop(self, key: str):
        assert key in list(
            self.keys()
        ), "KeyError: Cannot pop a tensor for a key that does not exist in the LazyStateDict."
        if self._state_dict_cache is not None and key in self._state_dict_cache:
            if key in self._index:
                self._index.pop(key)
            return self._state_dict_cache.pop(key)
        if key in self._index:
            self._index.pop(key)
        return None

    def __setitem__(self, key: str, value: torch.Tensor) -> None:
        """
        Set a tensor in the LazyStateDict. This will update the state dict cache if it is enabled.
        """
        assert key in list(
            self.keys()
        ), "KeyError: Cannot set a tensor for a key that does not exist in the LazyStateDict."
        if self._state_dict_cache is not None:
            self._state_dict_cache[key] = value
        else:
            log.warning("State dict cache is disabled, initializing the cache.")
            self._state_dict_cache = {key: value}

    def __contains__(self, key: str) -> bool:
        if self._state_dict_cache is not None and key in self._state_dict_cache:
            return True
        if self._index is not None and key in self._index:
            return True
        if len(self._checkpoint_files) == 1 and os.path.isfile(
            self._checkpoint_files[0]
        ):
            try:
                tensor = self._load_tensor_from_checkpoint_file(
                    self._checkpoint_files[0], key, update_cache=False
                )
                return tensor is not None
            except (KeyError, FileNotFoundError, RuntimeError, EOFError):
                return False
        return False

    def __len__(self) -> int:
        if self._index is not None:
            return len(self._index)
        if len(self._checkpoint_files) == 1 and os.path.isfile(
            self._checkpoint_files[0]
        ):
            checkpoint_file = self._checkpoint_files[0]
            if checkpoint_file.endswith(".safetensors"):
                with safe_open(checkpoint_file, framework="pt", device="cpu") as f:
                    return len(tuple(f.keys()))
            else:
                return len(
                    tuple(torch.load(checkpoint_file, map_location="cpu").keys())
                )
        raise RuntimeError(
            "Unexpected error: cannot determine the number of keys in the state dict."
        )

    def __iter__(self) -> Iterator[str]:
        if self._index is not None:
            return iter(self._index)
        elif self._state_dict_cache is not None:
            return iter(self._state_dict_cache)
        else:
            raise RuntimeError(
                "Unexpected error: cannot determine the keys in the state dict."
            )

    def keys(self) -> Iterator[str]:
        for key in self:
            yield key

    def values(self) -> Iterator[torch.Tensor]:
        for key in self:
            yield self[key]

    def items(self) -> Iterator[Tuple[str, torch.Tensor]]:
        for key in self:
            yield key, self[key]

    def __repr__(self) -> str:
        if self._index is not None:
            return f"{self.__class__.__name__}(keys={list(self.keys())})"
        else:
            return (
                f"{self.__class__.__name__}(checkpoint_files={self._checkpoint_files})"
            )

    def get_parameter(self, target: str) -> torch.Tensor:
        return self[target]

    def get_submodule(self, target: str) -> nn.Module:
        if self.meta_module is not None:
            module: nn.Module = deepcopy(self.meta_module.get_submodule(target))
            module.to_empty(device=self._device)
            state_dict = {}
            for name, _ in module.named_parameters():
                state_dict[name] = self[f"{target}.{name}"]
            module.load_state_dict(state_dict)
            return module
        else:
            raise RuntimeError(
                "Cannot get submodule because meta_module is not provided."
            )

    def load_state_dict(
        self, state_dict: Mapping[str, torch.Tensor], strict: bool = True
    ) -> _IncompatibleKeys:
        """
        Load a state dict into this LazyStateDict.
        This method is only for compatibility with nn.Module and it overrides the cache of LazyStateDict.

        Args:
            state_dict (Dict[str, torch.Tensor]): The state dict to load.
            strict (bool): Whether to enforce that all keys in the state dict are present in this LazyStateDict.
        """
        if not isinstance(state_dict, Mapping):
            raise TypeError(
                f"Expected state_dict to be dict-like, got {type(state_dict)}."
            )

        missing_keys: list[str] = []
        unexpected_keys: list[str] = []
        error_msgs: list[str] = []

        log.warning(
            "Loading state dict into LazyStateDict is not recommended, as it may lead to unexpected behavior. "
            "Use with caution."
        )

        # Check for unexpected keys in the provided state_dict
        for key in state_dict:
            if key not in self:
                unexpected_keys.append(key)

        # Check for missing keys that are expected in this LazyStateDict
        for key in self.keys():
            if key not in state_dict:
                missing_keys.append(key)

        # Handle strict mode
        if strict:
            if len(unexpected_keys) > 0:
                error_msgs.insert(
                    0,
                    "Unexpected key(s) in state_dict: {}. ".format(
                        ", ".join(f'"{k}"' for k in unexpected_keys)
                    ),
                )
            if len(missing_keys) > 0:
                error_msgs.insert(
                    0,
                    "Missing key(s) in state_dict: {}. ".format(
                        ", ".join(f'"{k}"' for k in missing_keys)
                    ),
                )

        if len(error_msgs) > 0:
            raise RuntimeError(
                "Error(s) in loading state_dict for {}:\n\t{}".format(
                    self.__class__.__name__, "\n\t".join(error_msgs)
                )
            )

        # Load the state dict values
        for key, value in state_dict.items():
            if key in self:  # Only set keys that exist in this LazyStateDict
                self[key] = value

        return _IncompatibleKeys(missing_keys, unexpected_keys)

    def __getattr__(self, name: str):
        if "meta_module" in self.__dict__:
            meta_module = self.__dict__["meta_module"]
            if meta_module is not None:
                if "_parameters" in meta_module.__dict__:
                    if name in meta_module.__dict__["_parameters"]:
                        return self.get_parameter(name)
                if "_modules" in meta_module.__dict__:
                    if name in meta_module.__dict__["_modules"]:
                        return self.get_submodule(name)
        raise AttributeError(
            f"'{type(self).__name__}' object has no attribute '{name}'"
        )



import os
from typing import Optional

from transformers import PreTrainedModel
from typing_extensions import override

from fusion_bench.modelpool.causal_lm.causal_lm import CausalLMPool
from fusion_bench.utils import timeit_context

from .depth_upscaling import DepthUpscalingAlgorithm


class DepthUpscalingForLlama(DepthUpscalingAlgorithm):
    """
    Implements depth upscaling for Llama models.

    This class extends the DepthUpscalingAlgorithm to handle Llama models specifically.
    It supports saving the upscaled model to a specified path.

    Args:
        layer_indices (list): List of layer indices to upscale.
        model_save_path (Optional[str]): Path to save the upscaled model.
        **kwargs: Additional keyword arguments.
    """

    def __init__(self, layer_indices: list, model_save_path: Optional[str], **kwargs):
        if isinstance(model_save_path, str):
            model_save_path = os.path.expanduser(model_save_path)
        self.model_save_path = model_save_path
        super().__init__(layer_indices, **kwargs)

    @override
    def run(self, modelpool: CausalLMPool):
        """
        Executes the depth upscaling algorithm on a given model pool.

        This method loads the pretrained model or the first model in the pool,
        applies the depth upscaling algorithm, and updates the number of hidden layers in the model configuration.
        If a save path is provided, it saves the upscaled model and tokenizer to the specified path.

        Args:
            modelpool (CausalLMPool): The pool of models to upscale.

        Returns:
            CausalLM: The upscaled model.
        """
        if self.model_save_path is not None:
            tokenizer = modelpool.load_tokenizer()

        model: PreTrainedModel = modelpool.load_pretrained_or_first_model()
        model.model.layers = super().run(model.model.layers)
        model.config.num_hidden_layers = len(model.model.layers)

        if self.model_save_path is not None:
            with timeit_context(f"Saving the model to {self.model_save_path}"):
                tokenizer.save_pretrained(self.model_save_path)
                model.save_pretrained(self.model_save_path)
        return model



import logging
import os
from typing import Any, Dict, List, Optional

from datasets import load_dataset, load_from_disk
from transformers import PreTrainedTokenizer

import fusion_bench

log = logging.getLogger(__name__)


def load_tokenized_wiki_dataset(
    tokenizer: Optional[PreTrainedTokenizer],
    path: str = "wikitext",
    name: str = "wikitext-2-raw-v1",
    split: Optional[str] = None,
    datasets: Optional[Any] = None,
    block_size: int = 128,
    cache_path: Optional[str] = None,
):
    """
    Reference: https://github.com/huggingface/notebooks/blob/main/examples/language_modeling.ipynb

    Args:
        block_size (int):
        dataset: If dataset is provided, `path` and `name` will be ignored.
    """
    if cache_path is not None and fusion_bench.utils.path.path_is_dir_and_not_empty(
        cache_path
    ):
        datasets = load_from_disk(cache_path)
        if split is None:
            return datasets
        else:
            return datasets[split]
    else:
        assert (
            tokenizer is not None
        ), "Cached dataset not found. Need tokenizer to process the raw data."

    # 1. load raw dataset
    if datasets is not None:
        log.info("Use `datasets`, `path` and `name` are ignored.")
    else:
        datasets = load_dataset(path, name)

    # 2. tokenize the dataset
    def tokenize_function(examples):
        return tokenizer(examples["text"])

    tokenized_datasets = datasets.map(
        tokenize_function, batched=True, num_proc=4, remove_columns=["text"]
    )

    # If we now look at an element of our datasets, we will see the text have been replaced by the input_ids the model will need:
    # { 'attention_mask': <list of int>, 'input_ids': <list of int> }

    # 3. concat and truncate tokens
    def group_texts(examples: Dict[str, List]):
        # Concatenate all texts.
        concatenated_examples = {k: sum(examples[k], []) for k in examples.keys()}
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        # We drop the small remainder, we could add padding if the model supported it instead of this drop, you can
        # customize this part to your needs.
        total_length = (total_length // block_size) * block_size
        # Split by chunks of max_len.
        result = {
            k: [t[i : i + block_size] for i in range(0, total_length, block_size)]
            for k, t in concatenated_examples.items()
        }
        result["labels"] = result["input_ids"].copy()
        return result

    lm_datasets = tokenized_datasets.map(
        group_texts,
        batched=True,
        batch_size=1000,
        num_proc=4,
    )

    if cache_path is not None:
        os.makedirs(cache_path, exist_ok=True)
        lm_datasets.save_to_disk(cache_path)

    if split is None:
        return lm_datasets
    else:
        return lm_datasets[split]



import logging
from typing import List, Mapping, Union  # noqa: F401

import numpy as np
import torch
from transformers import PreTrainedModel
from typing_extensions import override

from fusion_bench.method import BaseAlgorithm
from fusion_bench.mixins import auto_register_config
from fusion_bench.modelpool import CausalLMPool
from fusion_bench.utils import timeit_context
from fusion_bench.utils.state_dict_arithmetic import state_dict_add, state_dict_mul
from fusion_bench.utils.type import StateDictType

log = logging.getLogger(__name__)


@auto_register_config
class WeightedAverageForLLama(BaseAlgorithm):
    """
    A class to perform weighted averaging of LlaMa/Mistral models.
    """

    def __init__(
        self,
        normalize: bool,
        weights: List[float],
        backbone_only: bool,
        merged_model_save_path: str,
        save_tokenizer: bool,
        push_to_hub: bool,
        **kwargs,
    ):
        """
        Initialize the WeightedAverageForLLama class with the given parameters.

        Args:
            normalize (bool): Whether to normalize the weights.
            weights (List[float]): The weights for averaging the models.
            backbone_only (bool): Whether to use only the backbone of the models.
            merged_model_save_path (str): The path to save the merged model.
            save_tokenizer (bool): Whether to save the tokenizer.
            push_to_hub (bool): Whether to push the model to the hub.
        """
        super().__init__(**kwargs)

    @override
    @torch.no_grad()
    def run(self, modelpool: CausalLMPool) -> PreTrainedModel:
        """
        Executes the weighted averaging of models in the provided model pool.

        Args:
            modelpool (LLamaForCausalLMPoolThe):  pool of models to be averaged.

        Returns:
            base_model: The base model after merging the state dictionaries of the models in the pool.

        Raises:
            ValueError: If the number of weights does not match the number of models in the pool.
        """
        if modelpool.has_pretrained:
            base_model = modelpool.load_model("_pretrained_")
        else:
            base_model = modelpool.load_model(modelpool.model_names[0])

        weights = self.weights
        if len(weights) != len(modelpool.model_names):
            raise ValueError(
                "Number of weights must match the number of models.,"
                f"but got {len(weights)} weights and {len(modelpool.model_names)} models."
                f"weights: {weights}, models: {modelpool.model_names}"
            )
        if self.normalize:
            weights = np.asarray(weights)
            weights = weights / np.sum(weights)

        merged_state_dict: StateDictType = None
        for model_name, weight in zip(modelpool.model_names, weights):
            model = modelpool.load_model(model_name, backbone_only=self.backbone_only)
            sd = state_dict_mul(model.state_dict(), weight)
            if merged_state_dict is None:
                merged_state_dict = sd
            else:
                merged_state_dict = state_dict_add(merged_state_dict, sd)

        base_model.load_state_dict(
            merged_state_dict, strict=False if self.backbone_only else True
        )
        if self.merged_model_save_path is not None:
            with timeit_context(
                f"Saving the merged model to {self.merged_model_save_path}"
            ):
                modelpool.save_model(
                    base_model,
                    path=self.merged_model_save_path,
                    save_tokenizer=self.save_tokenizer,
                    push_to_hub=self.push_to_hub,
                )
        return base_model



from transformers import AutoModelForSequenceClassification


def create_reward_model_from_pretrained(pretrained_model_name_or_path: str, **kwargs):
    """
    Create a reward model for reward modeling (RLHF).

    Args:
        pretrained_model_name_or_path (str): The name or path of the pretrained model.
        **kwargs: Additional keyword arguments passed to the model class.
    """
    model = AutoModelForSequenceClassification.from_pretrained(
        pretrained_model_name_or_path, num_labels=1, **kwargs
    )
    return model



import itertools
import json
import logging
import os
from typing import TYPE_CHECKING, Callable, Dict, Optional, Union

import lightning.fabric
import open_clip
import torch
from omegaconf import DictConfig
from torch.nn import functional as F
from torch.utils.data import DataLoader, Dataset
from torchmetrics import Accuracy, MeanMetric
from torchmetrics.classification.accuracy import MulticlassAccuracy
from tqdm.auto import tqdm

from fusion_bench import BaseTaskPool
from fusion_bench.dataset.clip_dataset import CLIPDataset
from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.modelpool.openclip_vision.modelpool import load_classifier_head
from fusion_bench.models.open_clip import (
    ClassificationHead,
    ImageClassifier,
    ImageEncoder,
)
from fusion_bench.models.open_clip.variables_and_paths import OPENCLIP_CACHEDIR
from fusion_bench.utils import count_parameters, instantiate

if TYPE_CHECKING:
    from fusion_bench.modelpool import OpenCLIPVisionModelPool
    from fusion_bench.programs import FabricModelFusionProgram

log = logging.getLogger(__name__)


class OpenCLIPVisionModelTaskPool(
    BaseTaskPool,
    LightningFabricMixin,
):
    _is_setup = False

    _program: "FabricModelFusionProgram"

    processor: Optional[Callable] = None
    test_datasets: Dict[str, CLIPDataset]

    def __init__(
        self,
        test_datasets: Union[DictConfig, Dict[str, Dataset]],
        classification_heads: Union[DictConfig, Dict[str, ClassificationHead]],
        dataloader_kwargs: DictConfig,
        model_name: Optional[str] = None,
        fast_dev_run: bool = False,
        **kwargs,
    ):
        self._test_datasets = test_datasets
        self._classifier_heads = classification_heads
        self._dataloader_kwargs = dataloader_kwargs
        self._model_name = model_name
        self.fast_dev_run = fast_dev_run
        super().__init__(**kwargs)

    def setup(self):
        # setup the processor
        if self._program is not None and self._program.modelpool is not None:
            modelpool: "OpenCLIPVisionModelPool" = self._program.modelpool
            self.processor = modelpool.test_processor
        elif self._model_name is not None:
            _, _, self.processor = open_clip.create_model_and_transforms(
                self._model_name,
                pretrained="openai",
                cache_dir=OPENCLIP_CACHEDIR,
            )
        else:
            raise ValueError("Modelpool or model_name is not set")

        # setup the test datasets
        self.test_datasets = {
            name: instantiate(dataset) if isinstance(dataset, DictConfig) else dataset
            for name, dataset in self._test_datasets.items()
        }
        self.test_datasets = {
            name: CLIPDataset(dataset, self.processor)
            for name, dataset in self.test_datasets.items()
        }
        self.test_dataloaders = {
            name: self.fabric.setup_dataloaders(
                DataLoader(dataset, **self._dataloader_kwargs)
            )
            for name, dataset in self.test_datasets.items()
        }

        # setup classifier heads
        self.classifier_heads = {
            name: load_classifier_head(head).to(self.fabric.device)
            for name, head in self._classifier_heads.items()
        }
        self._is_setup = True

    @torch.no_grad()
    def _evaluate(
        self,
        classifier: ImageClassifier,
        test_loader: DataLoader,
        num_classes: int,
        task_name: str,
    ):
        accuracy: MulticlassAccuracy = Accuracy(
            task="multiclass", num_classes=num_classes
        )
        classifier.eval()
        loss_metric = MeanMetric()
        # if fast_dev_run is set, we only evaluate on a batch of the data
        if self.fast_dev_run:
            log.info("Running under fast_dev_run mode, evaluating on a single batch.")
            test_loader = itertools.islice(test_loader, 1)
        else:
            test_loader = test_loader

        pbar = tqdm(
            test_loader,
            desc=f"Evaluating {task_name}",
            leave=False,
            dynamic_ncols=True,
        )
        for batch in pbar:
            inputs, targets = batch
            logits = classifier(inputs)
            loss = F.cross_entropy(logits, targets)
            loss_metric.update(loss.detach().cpu())
            acc = accuracy(logits.detach().cpu(), targets.detach().cpu())
            pbar.set_postfix(
                {
                    "accuracy": accuracy.compute().item(),
                    "loss": loss_metric.compute().item(),
                }
            )

        acc = accuracy.compute().item()
        loss = loss_metric.compute().item()
        results = {"accuracy": acc, "loss": loss}
        return results

    def evaluate(self, model: ImageEncoder, **kwargs):
        if not self._is_setup:
            self.setup()

        report = {}
        # collect basic model information
        training_params, all_params = count_parameters(model)
        report["model_info"] = {
            "trainable_params": training_params,
            "all_params": all_params,
            "trainable_percentage": training_params / all_params,
        }

        if not lightning.fabric.is_wrapped(model):
            model = self.fabric.setup_module(model)

        pbar = tqdm(
            self.test_dataloaders.items(),
            desc="Evaluating tasks",
            total=len(self.test_dataloaders),
        )
        for task_name, test_dataloader in pbar:
            classifier = ImageClassifier(model, self.classifier_heads[task_name])
            num_classes = self.classifier_heads[task_name].weight.size(0)
            result = self._evaluate(
                classifier,
                test_dataloader,
                num_classes=num_classes,
                task_name=task_name,
            )
            report[task_name] = result

        # calculate the average accuracy and loss
        if "average" not in report:
            report["average"] = {}
            accuracies = [
                value["accuracy"]
                for key, value in report.items()
                if "accuracy" in value
            ]
            if len(accuracies) > 0:
                average_accuracy = sum(accuracies) / len(accuracies)
                report["average"]["accuracy"] = average_accuracy
            losses = [value["loss"] for key, value in report.items() if "loss" in value]
            if len(losses) > 0:
                average_loss = sum(losses) / len(losses)
                report["average"]["loss"] = average_loss

        log.info(f"Evaluation Result: {report}")
        if self.fabric.is_global_zero and len(self.fabric._loggers) > 0:
            with open(os.path.join(self.log_dir, "report.json"), "w") as fp:
                json.dump(report, fp)
        return report



import torch


def get_memory_usage(desc):
    """
    obtain the current GPU memory usage

    Returns:
        str: A string containing the allocated and cached memory in MB.
    """
    allocated = torch.cuda.memory_allocated() / 1024**2  # 转换为 MB
    cached = torch.cuda.memory_reserved() / 1024**2  # 转换为 MB
    return (
        f"{desc}\nAllocated Memory: {allocated:.2f} MB\nCached Memory: {cached:.2f} MB"
    )



import logging
from typing import Dict, List, Mapping, Optional, TypeVar, Union  # noqa: F401

from typing_extensions import override

from fusion_bench import timeit_context
from fusion_bench.method import TaskArithmeticAlgorithm
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.modelpool import CausalLMBackbonePool, CausalLMPool

log = logging.getLogger(__name__)


class TaskArithmeticForLlama(TaskArithmeticAlgorithm, SimpleProfilerMixin):
    R"""
    Examples:

    fusion_bench \
        method=linear/task_arithmetic_for_llama \
            method.scaling_factor=0.3 \
        method.model_save_path=outputs/simle_mixtral_exp_v4/task_arithmetic_0.3 \
        modelpool=CausalLMPool/simle_mixtral_exp_v4.yaml
    """

    _config_mapping = TaskArithmeticAlgorithm._config_mapping | {
        "merge_backbone": "merge_backbone",
    }

    def __init__(
        self,
        scaling_factor: float,
        merge_backbone: bool,
        model_save_path: Optional[str] = None,
    ):
        self.merge_backbone = merge_backbone
        self.model_save_path = model_save_path
        super().__init__(scaling_factor=scaling_factor)

    @override
    def run(self, modelpool: CausalLMPool):
        if self.model_save_path:
            tokenizer = modelpool.load_tokenizer()

        if self.merge_backbone:
            assert modelpool.has_pretrained
            backbone_modelpool = CausalLMBackbonePool(**modelpool.config)
            model = modelpool.load_model("_pretrained_")
            backbone_model = super().run(backbone_modelpool)
            model.model.layers = backbone_model
        else:
            model = super().run(modelpool)

        if self.model_save_path is not None:
            with timeit_context(f"Saving the model to {self.model_save_path}"):
                tokenizer.save_pretrained(self.model_save_path)
                model.save_pretrained(self.model_save_path)
        return model



from transformers import AutoConfig, AutoModel, AutoModelForCausalLM

from .configuration_losparse_llama import LoSparseLlamaConfig
from .modeling_losparse_llama import LoSparseLlamaForCausalLM, LoSparseLlamaModel

AutoConfig.register("losparse_llama", LoSparseLlamaConfig)
AutoModel.register(LoSparseLlamaConfig, LoSparseLlamaModel)
AutoModelForCausalLM.register(LoSparseLlamaConfig, LoSparseLlamaForCausalLM)



import numpy as np
from IPython.display import HTML, display


def create_color_style():
    return """
    <style>
        .token-container { font-family: monospace; white-space: pre; }
        .attention { background-color: #90EE90; }  /* Light green */
        .label { background-color: #FFB6C6; }      /* Light red */
        .token { color: #0066cc; }                 /* Blue */
        .stats { font-weight: bold; }
    </style>
    """


def escape_special_chars(text):
    """Convert special characters to their string representation"""
    return (
        text.replace("\n", "\\n")
        .replace("\t", "\\t")
        .replace("\r", "\\r")
        .replace(" ", "␣")
    )  # Optional: show spaces with visible character


def visualize_tokens_html(input_ids, attention_mask, labels, tokenizer):
    """
    Visualize model inputs using HTML colored text representation for Jupyter Notebook
    with special characters shown as strings
    """
    # Convert to numpy if tensors
    attention_mask = np.array(attention_mask).flatten()
    labels = np.array(labels).flatten()
    input_ids = np.array(input_ids).flatten()

    # Decode tokens and escape special characters
    tokens = [escape_special_chars(tokenizer.decode(id_)) for id_ in input_ids]

    # Create HTML output
    html_output = [create_color_style()]

    # Header
    html_output.append("<h3>**Token Visualization**</h3>")

    # Legend
    html_output.append(
        """
    <div style='margin: 10px 0;'>
        <strong>Legend:</strong><br>
        <span class='attention'>&nbsp;&nbsp;&nbsp;&nbsp;</span> Active Attention<br>
        <span class='label'>&nbsp;&nbsp;&nbsp;&nbsp;</span> Label Present<br>
        <span class='token'>Text</span> Token Text<br>
        Special Characters: \\n (newline), \\t (tab), ␣ (space)
    </div>
    """
    )

    # Token alignment
    html_output.append("<strong>Token Alignment:</strong>")
    html_output.append("<div class='token-container'>")

    # Calculate maximum token length for better alignment
    max_token_len = max(len(str(token)) for token in tokens)

    for i, (input_id, token, mask, label) in enumerate(
        zip(input_ids, tokens, attention_mask, labels)
    ):
        # Pad token for alignment
        token_text = f"{token:{max_token_len}s}"

        # Create classes for styling
        classes = []
        if mask == 1:
            classes.append("attention")
        if label != -100 and label != 0:
            classes.append("label")

        class_str = f"class='{' '.join(classes)}'" if classes else ""

        # Create the line
        line = f"Position {i:3d}: <span {class_str}><span class='token'>{token_text}</span></span> "
        line += (
            f"(Mask: {int(mask)}, Label: {int(label)}, Inpu_id: {int(input_id)})<br>"
        )
        html_output.append(line)

    html_output.append("</div>")

    # Statistics
    html_output.append(
        """
    <div class='stats' style='margin-top: 10px;'>
        Statistics:<br>
        Total tokens: {}<br>
        Active attention tokens: {}<br>
        Labeled tokens: {}<br>
    </div>
    """.format(
            len(tokens), attention_mask.sum(), sum(labels != -100)
        )
    )

    # Display the HTML
    display(HTML("".join(html_output)))


# Example usage:
"""
from transformers import AutoTokenizer
import torch

# Initialize tokenizer
tokenizer = AutoTokenizer.from_pretrained('bert-base-uncased')

# Sample input with special characters
text = "Hello,\nhow are\tyou?"
inputs = tokenizer(text, return_tensors='pt')
labels = torch.zeros_like(inputs['input_ids'])  # dummy labels

visualize_tokens_html(
    inputs['attention_mask'][0],
    labels[0],
    inputs['input_ids'][0],
    tokenizer
)
"""



class NoSparseGradientError(Exception):
    """Raised when the gradient is sparse gradient.

    :param optimizer_name: str. optimizer name.
    :param note: str. special conditions to note (default '').
    """

    def __init__(self, optimizer_name: str, note: str = ""):
        self.note: str = " " if not note else f" w/ {note} "
        self.message: str = (
            f"[-] {optimizer_name}{self.note}does not support sparse gradient."
        )
        super().__init__(self.message)


class ZeroParameterSizeError(Exception):
    """Raised when the parameter size is 0."""

    def __init__(self):
        self.message: str = "[-] parameter size is 0"
        super().__init__(self.message)


class NoClosureError(Exception):
    """Raised when there's no closure function."""

    def __init__(self, optimizer_name: str, note: str = ""):
        self.message: str = f"[-] {optimizer_name} requires closure.{note}"
        super().__init__(self.message)


class NegativeLRError(Exception):
    """Raised when learning rate is negative."""

    def __init__(self, lr: float, lr_type: str = ""):
        self.note: str = lr_type if lr_type else "learning rate"
        self.message: str = f"[-] {self.note} must be positive. ({lr} > 0)"
        super().__init__(self.message)


class NegativeStepError(Exception):
    """Raised when step is negative."""

    def __init__(self, num_steps: int, step_type: str = ""):
        self.note: str = step_type if step_type else "step"
        self.message: str = f"[-] {self.note} must be positive. ({num_steps} > 0)"
        super().__init__(self.message)



import logging
import os
import re
from copy import deepcopy
from typing import Dict, List, Tuple  # noqa: F401

import torch
from torch import Tensor, nn
from tqdm.auto import tqdm

from fusion_bench.compat.method import ModelFusionAlgorithm
from fusion_bench.compat.modelpool import ModelPool, to_modelpool
from fusion_bench.mixins.simple_profiler import SimpleProfilerMixin
from fusion_bench.models.utils import get_attr, set_attr

log = logging.getLogger(__name__)


def svd(w: Tensor, full_matrices: bool) -> Tuple[Tensor, Tensor, Tensor]:
    """
    Perform Singular Value Decomposition (SVD) on the given tensor.

    Args:
        w (Tensor): The input tensor to decompose.
        full_matrices (bool): Whether to compute the full-sized U and V matrices.

    Returns:
        Tuple[Tensor, Tensor, Tensor]: The U, S, and V matrices from the SVD.
    """
    u, s, vh = torch.linalg.svd(
        w, full_matrices=full_matrices, driver="gesvd" if w.is_cuda else None
    )
    v = vh.T
    return u, s, v


def _is_name_matched(name: str, extract_names: List[str]) -> bool:
    """
    Check if the given name matches any of the provided regular expressions.

    Args:
        name (str): The name to check.
        extract_names (List[str]): A list of regular expressions to match against.

    Returns:
        bool: True if the name matches any of the regular expressions, False otherwise.
    """
    for extract_name in extract_names:
        # extract_name is a regular expression
        if re.match(extract_name, name):
            return True
    return False


def _total_parameters(state) -> int:
    """
    Calculate the total number of parameters in the given state.

    Args:
        state: The state to calculate the parameters for. Can be a Tensor or a dictionary of Tensors.

    Returns:
        int: The total number of parameters.

    Raises:
        ValueError: If the state is not a Tensor or a dictionary of Tensors.
    """
    if isinstance(state, Tensor):
        return state.numel()
    elif isinstance(state, dict):
        return sum(_total_parameters(v) for v in state.values())
    else:
        raise ValueError(f"Unsupported type: {type(state)}")


class SingularProjectionMergingAlgorithm(ModelFusionAlgorithm, SimpleProfilerMixin):
    """
    A model fusion algorithm that projects parameter differences into the SVD subspace of a pretrained model.

    This algorithm is experimental and aims to investigate the location of task-specific knowledge.
    """

    @torch.no_grad()
    def run(self, modelpool: ModelPool) -> nn.Module:
        """
        Run the singular projection merging algorithm on the given model pool.

        Args:
            modelpool (ModelPool): The pool of models to merge.

        Returns:
            nn.Module: The merged model.
        """
        modelpool = to_modelpool(modelpool)

        if self.config.model_path is not None and os.path.exists(
            self.config.model_path
        ):
            log.info(f"loading merged model from {self.config.model_path}")
            model = torch.load(self.config.model_path)

        with self.profile("load pretrained model"):
            pretrained_model = modelpool.load_model("_pretrained_").to(
                self.config.device
            )
        with self.profile("load fine-tuned model"):
            finetuned_models = modelpool.load_model(modelpool.model_names[0]).to(
                self.config.device
            )

        with self.profile("merge model"):
            model = self.merge(pretrained_model, finetuned_models)

        if self.config.model_path is not None:
            os.path.makedirs(os.path.dirname(self.config.model_path), exist_ok=True)
            torch.save(model, self.config.model_path)

        self.print_profile_summary()
        return model

    def merge(
        self,
        pretrained_model: nn.Module,
        finetuned_model: nn.Module,
        in_place: bool = True,
    ) -> nn.Module:
        """
        Merges the pretrained model with the fine-tuned model by projecting parameter differences
        into the SVD subspace of the pretrained model.

        Args:
            pretrained_model (nn.Module): The pretrained model.
            finetuned_model (nn.Module): The fine-tuned model.
            in_place (bool): If True, modifies the fine-tuned model in place. Otherwise, creates a copy.

        Returns:
            nn.Module: The merged model.
        """
        if in_place:
            model = finetuned_model
        else:
            model = deepcopy(finetuned_model)

        for name, module in tqdm(
            tuple(model.named_modules()),
            "Projection merging in SVD subspace of pretrained model",
        ):
            if isinstance(module, nn.Linear):
                name_list = name.split(".")
                set_attr(
                    model,
                    name_list,
                    self.projection_merge_linear(
                        get_attr(pretrained_model, name_list),
                        get_attr(finetuned_model, name_list),
                        k=self.config.k,
                    ),
                )
        return model

    def projection_merge_linear(
        self, pretrained_model: nn.Linear, finetuned_model: nn.Linear, k: int
    ) -> nn.Linear:
        """
        Projects the parameter differences of linear layers into the SVD subspace of the pretrained model.

        Args:
            pretrained_model (nn.Linear): The linear layer of the pretrained model.
            finetuned_model (nn.Linear): The linear layer of the fine-tuned model.
            k (int): The number of singular values to keep. If negative, it is determined based on the sum of singular values.

        Returns:
            nn.Linear: The merged linear layer with projected parameter differences.
        """
        w = pretrained_model.weight
        w_ft = finetuned_model.weight

        u, s, v = svd(w, full_matrices=self.config.full_matrices)
        if k < 0:
            # find the position where the sum of singular values is larger than 50% of the total sum
            cumsum = s.cumsum(0)
            k = (cumsum < cumsum[-1] * 0.5).sum().item() + 1

        if self.config.rank == "low":
            u = u[:, :k]
            s = s[:k]
            v = v[:, :k]
        else:
            u = u[:, k:]
            s = s[k:]
            v = v[:, k:]

        w_diff = w_ft - w
        w_diff_proj = u.T @ w_diff @ v
        w.data = w + u @ w_diff_proj @ v.T
        if pretrained_model.bias is not None:
            pretrained_model.bias.data = finetuned_model.bias.data
        return pretrained_model



# ███████╗██╗   ██╗███████╗██╗ ██████╗ ███╗   ██╗      ██████╗ ███████╗███╗   ██╗ ██████╗██╗  ██╗
# ██╔════╝██║   ██║██╔════╝██║██╔═══██╗████╗  ██║      ██╔══██╗██╔════╝████╗  ██║██╔════╝██║  ██║
# █████╗  ██║   ██║███████╗██║██║   ██║██╔██╗ ██║█████╗██████╔╝█████╗  ██╔██╗ ██║██║     ███████║
# ██╔══╝  ██║   ██║╚════██║██║██║   ██║██║╚██╗██║╚════╝██╔══██╗██╔══╝  ██║╚██╗██║██║     ██╔══██║
# ██║     ╚██████╔╝███████║██║╚██████╔╝██║ ╚████║      ██████╔╝███████╗██║ ╚████║╚██████╗██║  ██║
# ╚═╝      ╚═════╝ ╚══════╝╚═╝ ╚═════╝ ╚═╝  ╚═══╝      ╚═════╝ ╚══════╝╚═╝  ╚═══╝ ╚═════╝╚═╝  ╚═╝
# flake8: noqa: F401
from . import (
    constants,
    dataset,
    method,
    metrics,
    mixins,
    modelpool,
    models,
    optim,
    programs,
    taskpool,
    tasks,
    utils,
)
from .constants import RuntimeConstants
from .method import BaseAlgorithm, BaseModelFusionAlgorithm
from .mixins import auto_register_config
from .modelpool import BaseModelPool
from .models import (
    create_default_model_card,
    load_model_card_template,
    save_pretrained_with_remote_code,
    separate_io,
)
from .programs import BaseHydraProgram
from .taskpool import BaseTaskPool
from .utils import (
    BoolStateDictType,
    LazyStateDict,
    StateDictType,
    TorchModelType,
    cache_with_joblib,
    get_rankzero_logger,
    import_object,
    instantiate,
    parse_dtype,
    print_parameters,
    seed_everything_by_time,
    set_default_cache_dir,
    set_print_function_call,
    set_print_function_call_permeanent,
    timeit_context,
)



import logging

from fusion_bench.mixins import LightningFabricMixin
from fusion_bench.models.open_clip import ImageClassifier, ImageEncoder

log = logging.getLogger(__name__)


class OpenCLIPClassificationMixin(LightningFabricMixin):
    _train_processor = None
    _test_processor = None



import logging
from typing import Optional

from omegaconf import DictConfig

from fusion_bench.programs import BaseHydraProgram

log = logging.getLogger(__name__)


class GreetingProgram(BaseHydraProgram):
    """
    A simple program that greets users with a custom message.
    """

    _config_mapping = BaseHydraProgram._config_mapping | {
        "message": "message",
        "name": "name",
        "repeat_count": "repeat_count",
    }

    def __init__(
        self,
        message: str = "Hello",
        name: str = "World",
        repeat_count: int = 1,
        **kwargs,
    ):
        self.message = message
        self.name = name
        self.repeat_count = repeat_count
        super().__init__(**kwargs)

    def run(self):
        """Execute the greeting workflow."""
        log.info("Starting greeting program")

        # Create the greeting
        greeting = f"{self.message}, {self.name}!"

        # Print the greeting multiple times
        for i in range(self.repeat_count):
            if self.repeat_count > 1:
                print(f"[{i+1}/{self.repeat_count}] {greeting}")
            else:
                print(greeting)

        log.info("Greeting program completed")
        return greeting
